# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os
import logging
import time

from .util import LoggerWrapper, UpdateNProc, CopyConfig, MultiProcess, SetupConfigRNG
from .util import RetryIO, SetDefaultExt
from .value import ParseValue, CheckAllParams
from .input import ProcessInput
from .extra import valid_extra_outputs, SetupExtraOutput, WriteExtraOutputs
from .extra import AddExtraOutputHDUs, CheckNoExtraOutputHDUs
from .image import BuildImage, GetNObjForImage
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..utilities import ensure_dir
from ..fits import writeMulti

# This file handles building the output files according to the specifications in config['output'].
# This file includes the basic functionality, but it calls out to helper functions for the
# different types of output files.  It includes the implementation of the default output type,
# 'Fits'.  See output_multifits.py for 'MultiFits' and output_datacube.py for 'DataCube'.

# This module-level dict will store all the registered output types.
# See the RegisterOutputType function at the end of this file.
# The keys are the (string) names of the output types, and the values will be builder objects
# that will perform the different stages of processing to construct and write the output file(s).
valid_output_types = {}

def BuildFiles(nfiles, config, file_num=0, logger=None, except_abort=False):
    """
    Build a number of output files as specified in config.

    Parameters:
        nfiles:         The number of files to build.
        config:         A configuration dict.
        file_num:       If given, the first file_num. [default: 0]
        logger:         If given, a logger object to log progress. [default: None]
        except_abort:   Whether to abort processing when a file raises an exception (True)
                        or just report errors and continue on (False). [default: False]

    Returns:
        the final config dict that was used.
    """
    logger = LoggerWrapper(logger)
    t1 = time.time()

    # The next line relies on getting errors when the rng is undefined.  However, the default
    # rng is None, which is a valid thing to construct a Deviate object from.  So for now,
    # set the rng to object() to make sure we get errors where we are expecting to.
    config['rng'] = object()

    # Process the input field for the first file.  Often there are "safe" input items
    # that won't need to be reprocessed each time.  So do them here once and keep them
    # in the config for all file_nums.  This is more important if nproc != 1.
    ProcessInput(config, logger=logger, safe_only=True)

    jobs = []  # Will be a list of the kwargs to use for each job
    info = []  # Will be a list of (file_num, file_name) correspongind to each jobs.

    # Count from 0 to make sure image_num, etc. get counted right.  We'll start actually
    # building the files at first_file_num.
    first_file_num = file_num
    file_num = 0
    image_num = 0
    obj_num = 0

    # Figure out how many processes we will use for building the files.
    if 'output' not in config: config['output'] = {}
    output = config['output']
    if nfiles > 1 and 'nproc' in output:
        nproc = ParseValue(output, 'nproc', config, int)[0]
        # Update this in case the config value is -1
        nproc = UpdateNProc(nproc, nfiles, config, logger)
        # We'll want a pristine version later to give to the workers.
    else:
        nproc = 1
    orig_config = CopyConfig(config)

    if 'timeout' in output:
        timeout = ParseValue(output, 'timeout', config, float)[0]
    else:
        timeout = 3600

    if nfiles == 0:
        logger.warning("No files were made, since nfiles == 0.")
        return orig_config

    for k in range(nfiles + first_file_num):
        SetupConfigFileNum(config, file_num, image_num, obj_num, logger)

        builder = valid_output_types[output['type']]
        builder.setup(output, config, file_num, logger)

        # Process the input fields that might be relevant at file scope:
        ProcessInput(config, logger=logger, file_scope_only=True)

        # Get the number of objects in each image for this file.
        nobj = GetNObjForFile(config, file_num, image_num, logger=logger, approx=True)

        # The kwargs to pass to BuildFile
        kwargs = {
            'file_num' : file_num,
            'image_num' : image_num,
            'obj_num' : obj_num
        }

        if file_num >= first_file_num:
            # Get the file_name here, in case it needs to create directories, which is not
            # safe to do with multiple processes. (At least not without extra code in the
            # getFilename function...)
            file_name = builder.getFilename(output, config, logger)
            jobs.append(kwargs)
            info.append( (file_num, file_name) )

        # nobj is a list of nobj for each image in that file.
        # So len(nobj) = nimages and sum(nobj) is the total number of objects
        # This gets the values of image_num and obj_num ready for the next loop.
        file_num += 1
        image_num += len(nobj)
        obj_num += sum(nobj)

    def done_func(logger, proc, k, result, t2):
        file_num, file_name = info[k]
        file_name2, t = result  # This is the t for which 0 means the file was skipped.
        if file_name2 != file_name:  # pragma: no cover  (I think this should never happen.)
            raise GalSimConfigError("Files seem to be out of sync. %s != %s", file_name, file_name2)
        if t != 0 and logger:
            if proc is None: s0 = ''
            else: s0 = '%s: '%proc
            logger.warning(s0 + 'File %d = %s: time = %f sec', file_num, file_name, t)

    def except_func(logger, proc, k, e, tr):
        file_num, file_name = info[k]
        if proc is None: s0 = ''
        else: s0 = '%s: '%proc
        logger.error(s0 + 'Exception caught for file %d = %s', file_num, file_name)
        if except_abort:
            logger.debug('%s',tr)
            logger.error('File %s not written.',file_name)
        else:
            logger.warning('%s',tr)
            logger.error('File %s not written! Continuing on...',file_name)

    # Convert to the tasks structure we need for MultiProcess
    # Each task is a list of (job, k) tuples.  In this case, we only have one job per task.
    tasks = [ [ (job, k) ] for (k, job) in enumerate(jobs) ]

    results = MultiProcess(nproc, orig_config, BuildFile, tasks, 'file',
                           logger=logger, timeout=timeout,
                           done_func=done_func, except_func=except_func,
                           except_abort=except_abort)
    t2 = time.time()

    if len(results) == 0:
        nfiles_written = 0
    else:
        fnames, times = zip(*results)
        nfiles_written = sum([ t!=0 for t in times])

    if nfiles_written == 0:
        logger.warning('No files were written.  All were either skipped or had errors.')
    else:
        if nfiles_written > 1 and nproc != 1:
            logger.warning('Total time for %d files with %d processes = %f sec',
                           nfiles_written,nproc,t2-t1)
        logger.warning('Done building files')

    #Return the config used for the run - this may be useful since one can
    #save information here in e.g. custom output types
    return orig_config

output_ignore = [ 'nproc', 'timeout', 'skip', 'noclobber', 'retry_io' ]

def BuildFile(config, file_num=0, image_num=0, obj_num=0, logger=None):
    """
    Build an output file as specified in config.

    Parameters:
        config:         A configuration dict.
        file_num:       If given, the current file_num. [default: 0]
        image_num:      If given, the current image_num. [default: 0]
        obj_num:        If given, the current obj_num. [default: 0]
        logger:         If given, a logger object to log progress. [default: None]

    Returns:
        (file_name, t), a tuple of the file name and the time taken to build file
        Note: t==0 indicates that this file was skipped.
    """
    logger = LoggerWrapper(logger)
    t1 = time.time()

    SetupConfigFileNum(config, file_num, image_num, obj_num, logger)
    output = config['output']
    output_type = output['type']
    builder = valid_output_types[output_type]
    builder.setup(output, config, file_num, logger)

    # Put these values in the config dict so we won't have to run them again later if
    # we need them.  e.g. ExtraOuput processing uses these.
    nobj = GetNObjForFile(config, file_num, image_num, logger=logger)
    nimages = len(nobj)
    config['nimages'] = nimages
    config['nobj'] = nobj
    logger.debug('file %d: BuildFile with type=%s to build %d images, starting with %d',
                 file_num,output_type,nimages,image_num)

    # Make sure the inputs and extra outputs are set up properly.
    ProcessInput(config, logger=logger)
    SetupExtraOutput(config, logger=logger)

    # Get the file name
    file_name = builder.getFilename(output, config, logger)

    # Check if we ought to skip this file
    if 'skip' in output and ParseValue(output, 'skip', config, bool)[0]:
        logger.warning('Skipping file %d = %s because output.skip = True',file_num,file_name)
        return file_name, 0  # Note: time=0 is the indicator that a file was skipped.
    if ('noclobber' in output
        and ParseValue(output, 'noclobber', config, bool)[0]
        and os.path.isfile(file_name)):
        logger.warning('Skipping file %d = %s because output.noclobber = True'
                       ' and file exists',file_num,file_name)
        return file_name, 0

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('file %d: file_name = %s',file_num,file_name)
    else:
        logger.warning('Start file %d = %s', file_num, file_name)

    ignore = output_ignore + list(valid_extra_outputs)
    data = builder.buildImages(output, config, file_num, image_num, obj_num, ignore, logger)

    # If any images came back as None, then remove them, since they cannot be written.
    data = [ im for im in data if im is not None ]

    if len(data) == 0:
        logger.warning('Skipping file %d = %s because all images were None',file_num,file_name)
        return file_name, 0

    # Go back to file_num as the default index_key.
    config['index_key'] = 'file_num'

    data = builder.addExtraOutputHDUs(config, data, logger)

    if 'retry_io' in output:
        ntries = ParseValue(output,'retry_io',config,int)[0]
        # This is how many _re_-tries.  Do at least 1, so ntries is 1 more than this.
        ntries = ntries + 1
    else:
        ntries = 1

    args = (data, file_name, output, config, logger)
    RetryIO(builder.writeFile, args, ntries, file_name, logger)
    logger.debug('file %d: Wrote %s to file %r',file_num,output_type,file_name)

    builder.writeExtraOutputs(config, data, logger)

    t2 = time.time()

    return file_name, t2-t1

def GetNFiles(config, logger=None):
    """
    Get the number of files that will be made, based on the information in the config dict.

    Parameters:
        config:     The configuration dict.
        logger:     If given, a logger object to log progress. [default: None]

    Returns:
        the number of files
    """
    output = config.get('output',{})
    output_type = output.get('type','Fits')
    if output_type not in valid_output_types:
        raise GalSimConfigValueError("Invalid output.type.", output_type,
                                     list(valid_output_types.keys()))
    return valid_output_types[output_type].getNFiles(output, config, logger=logger)


def GetNImagesForFile(config, file_num, logger=None):
    """
    Get the number of images that will be made for the file number file_num, based on the
    information in the config dict.

    Parameters:
        config:     The configuration dict.
        file_num:   The current file number.
        logger:     If given, a logger object to log progress. [default: None]

    Returns:
        the number of images
    """
    output = config.get('output',{})
    output_type = output.get('type','Fits')
    if output_type not in valid_output_types:
        raise GalSimConfigValueError("Invalid output.type.", output_type,
                                     list(valid_output_types.keys()))
    return valid_output_types[output_type].getNImages(output, config, file_num, logger=logger)


def GetNObjForFile(config, file_num, image_num, logger=None, approx=False):
    """
    Get the number of objects that will be made for each image built as part of the file file_num,
    which starts at image number image_num, based on the information in the config dict.

    Parameters:
        config:         The configuration dict.
        file_num:       The current file number.
        image_num:      The current image number.
        logger:         If given, a logger object to log progress. [default: None]
        approx:         Whether an approximate/overestimate is ok [default: False]

    Returns:
        a list of the number of objects in each image [ nobj0, nobj1, nobj2, ... ]
    """
    output = config.get('output',{})
    output_type = output.get('type','Fits')
    if output_type not in valid_output_types:
        raise GalSimConfigValueError("Invalid output.type.", output_type,
                                     list(valid_output_types.keys()))
    return valid_output_types[output_type].getNObjPerImage(output, config, file_num, image_num,
                                                           logger=logger, approx=approx)


def SetupConfigFileNum(config, file_num, image_num, obj_num, logger=None):
    """Do the basic setup of the config dict at the file processing level.

    Includes:
    - Set config['file_num'] = file_num
    - Set config['image_num'] = image_num
    - Set config['obj_num'] = obj_num
    - Set config['index_key'] = 'file_num'
    - Set config['start_image_num'] = image_num
    - Set config['start_obj_num'] = obj_num
    - Make sure config['output'] exists
    - Set default config['output']['type'] to 'Fits' if not specified
    - Check that the specified output type is valid.

    Parameters:
        config:     A configuration dict.
        file_num:   The current file_num. (If file_num=None, then don't set file_num or
                    start_obj_num items in the config dict.)
        image_num:  The current image_num.
        obj_num:    The current obj_num.
        logger:     If given, a logger object to log progress. [default: None]
    """
    logger = LoggerWrapper(logger)
    config['file_num'] = file_num
    config['start_obj_num'] = obj_num
    config['start_image_num'] = image_num
    config['image_num'] = image_num
    config['obj_num'] = obj_num
    config['index_key'] = 'file_num'

    if 'output' not in config:
        config['output'] = {}
    if 'type' not in config['output']:
        config['output']['type'] = 'Fits'

    # Check that the type is valid
    output_type = config['output']['type']
    if output_type not in valid_output_types:
        raise GalSimConfigValueError("Invalid output.type.", output_type,
                                     list(valid_output_types.keys()))


class OutputBuilder:
    """A base class for building and writing the output objects.

    The base class defines the call signatures of the methods that any derived class should follow.
    It also includes the implementation of the default output type: Fits.
    """

    # A class attribute that sub-classes may override.
    default_ext = '.fits'

    def setup(self, config, base, file_num, logger):
        """Do any necessary setup at the start of processing a file.

        The base class just calls SetupConfigRNG, but this provides a hook for sub-classes to
        do more things before any processing gets started on this file.

        Parameters:
            config:     The configuration dict for the output type.
            base:       The base configuration dict.
            file_num:   The current file_num.
            logger:     If given, a logger object to log progress.
        """
        seed = SetupConfigRNG(base, logger=logger)
        logger.debug('file %d: seed = %d',file_num,seed)

    def getFilename(self, config, base, logger):
        """Get the file_name for the current file being worked on.

        Note that the base class defines a default extension = '.fits'.
        This can be overridden by subclasses by changing the default_ext property.

        Parameters:
            config:     The configuration dict for the output type.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress.

        Returns:
            the filename to build.
        """
        if 'file_name' in config:
            SetDefaultExt(config['file_name'], self.default_ext)
            file_name = ParseValue(config, 'file_name', base, str)[0]
        elif 'root' in base and self.default_ext is not None:
            # If a file_name isn't specified, we use the name of the config file + '.fits'
            file_name = base['root'] + self.default_ext
        else:
            raise GalSimConfigError(
                "No file_name specified and unable to generate it automatically.")

        # Prepend a dir to the beginning of the filename if requested.
        if 'dir' in config:
            dir = ParseValue(config, 'dir', base, str)[0]
            file_name = os.path.join(dir,file_name)

        ensure_dir(file_name)

        return file_name

    def buildImages(self, config, base, file_num, image_num, obj_num, ignore, logger):
        """Build the images for output.

        In the base class, this function just calls BuildImage to build the single image to
        put in the output file.  So the returned list only has one item.

        Parameters:
            config:     The configuration dict for the output field.
            base:       The base configuration dict.
            file_num:   The current file_num.
            image_num:  The current image_num.
            obj_num:    The current obj_num.
            ignore:     A list of parameters that are allowed to be in config that we can
                        ignore here.  i.e. it won't be an error if they are present.
            logger:     If given, a logger object to log progress.

        Returns:
            a list of the images built
        """
        # There are no extra parameters to get, so just check that there are no invalid parameters
        # in the config dict.
        ignore += [ 'file_name', 'dir', 'nfiles' ]
        CheckAllParams(config, ignore=ignore)

        image = BuildImage(base, image_num, obj_num, logger=logger)
        return [ image ]

    def getNFiles(self, config, base, logger=None):
        """Returns the number of files to be built.

        In the base class, this is just output.nfiles.

        Parameters:
            config:     The configuration dict for the output field.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress.

        Returns:
            the number of files to build.
        """
        if 'nfiles' in config:
            return ParseValue(config, 'nfiles', base, int)[0]
        else:
            return 1

    def getNImages(self, config, base, file_num, logger=None):
        """Returns the number of images to be built for a given ``file_num``.

        In the base class, we only build a single image, so it returns 1.

        Parameters:
            config:     The configuration dict for the output field.
            base:       The base configuration dict.
            file_num:   The current file number.
            logger:     If given, a logger object to log progress.

        Returns:
           the number of images to build.
        """
        return 1

    def getNObjPerImage(self, config, base, file_num, image_num, logger=None, approx=False):
        """
        Get the number of objects that will be made for each image built as part of the file
        file_num, which starts at image number image_num, based on the information in the config
        dict.

        Parameters:
            config:         The configuration dict.
            base:           The base configuration dict.
            file_num:       The current file number.
            image_num:      The current image number (the first one for this file).
            logger:         If given, a logger object to log progress.
            approx:         Whether an approximate/overestimate is ok [default: False]

        Returns:
            a list of the number of objects in each image [ nobj0, nobj1, nobj2, ... ]
        """
        nimages = self.getNImages(config, base, file_num, logger=logger)
        nobj = [ GetNObjForImage(base, image_num+j, logger=logger, approx=approx)
                 for j in range(nimages) ]
        base['image_num'] = image_num  # Make sure this is set back to current image num.
        return nobj

    def canAddHdus(self):
        """Returns whether it is permissible to add extra HDUs to the end of the data list.

        In the base class, this returns True.
        """
        return True

    def addExtraOutputHDUs(self, config, data, logger):
        """If appropriate, add any extra output items that go into HDUs to the data list.

        Parameters:
            config:     The configuration dict for the output field.
            data:       The data to write.  Usually a list of images.
            logger:     If given, a logger object to log progress.

        Returns:
            data (possibly updated with additional items)
        """
        if self.canAddHdus():
            data = AddExtraOutputHDUs(config, data, logger)
        else:
            CheckNoExtraOutputHDUs(config, config['output']['type'], logger)
        return data

    def writeFile(self, data, file_name, config, base, logger):
        """Write the data to a file.

        Parameters:
            data:           The data to write.  Usually a list of images returned by
                            buildImages, but possibly with extra HDUs tacked onto the end
                            from the extra output items.
            file_name:      The file_name to write to.
            config:         The configuration dict for the output field.
            base:           The base configuration dict.
            logger:         If given, a logger object to log progress.
        """
        writeMulti(data,file_name)

    def writeExtraOutputs(self, config, data, logger):
        """If appropriate, write any extra output items that write their own files.

        Parameters:
            config:     The configuration dict for the output field.
            data:       The data to write.  Usually a list of images.
            logger:     If given, a logger object to log progress.
        """
        WriteExtraOutputs(config, data, logger)


def RegisterOutputType(output_type, builder):
    """Register an output type for use by the config apparatus.

    Parameters:
        output_type:    The name of the type in config['output']
        builder:        A builder object to use for building and writing the output file.
                        It should be an instance of OutputBuilder or a subclass thereof.
    """
    # Make a concrete instance of the builder.
    valid_output_types[output_type] = builder

# The base class is also the builder for type = Fits.
RegisterOutputType('Fits', OutputBuilder())




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from .input import InputLoader, GetInputObj, RegisterInputType
from .value import GetAllParams, RegisterValueType
from ..fits import FitsHeader

# This file adds input type fits_header and value type FitsHeader.

def _GenerateFromFitsHeader(config, base, value_type):
    """Return a value read from a FITS header
    """
    header = GetInputObj('fits_header', config, base, 'FitsHeader')

    req = { 'key' : str }
    opt = { 'num' : int }
    kwargs, safe = GetAllParams(config, base, req=req, opt=opt)
    key = kwargs['key']

    val = header.get(key)

    #print(base['file_num'],'Header: key = %s, val = %s'%(key,val))
    return val, safe

# Register this as a valid value type
RegisterValueType('FitsHeader', _GenerateFromFitsHeader, [ float, int, bool, str ],
                  input_type='fits_header')

# The FitsHeader doesn't need anything special other than registration as a valid input type.
RegisterInputType('fits_header', InputLoader(FitsHeader, file_scope=True))

# Registering this after FitsHeader rather than above as I normally would is just a gratuitous
# test coverage edit to help cover the different branches in the RegisterInputType and
# RegisterConnectedInputType functions.




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

# Define the class hierarchy for errors and warnings emitted by GalSim that aren't
# obviously one of the standard python errors.

__all__ = [ 'GalSimError', 'GalSimValueError', 'GalSimKeyError', 'GalSimIndexError',
            'GalSimRangeError', 'GalSimBoundsError', 'GalSimUndefinedBoundsError',
            'GalSimImmutableError', 'GalSimIncompatibleValuesError',
            'GalSimSEDError', 'GalSimHSMError', 'GalSimFFTSizeError',
            'GalSimConfigError', 'GalSimConfigValueError',
            'GalSimNotImplementedError',
            'GalSimWarning', 'GalSimDeprecationWarning', 'GalSimFFTSizeWarning',
            'convert_cpp_errors', 'galsim_warn', 'galsim_warn_fft' ]

import warnings
from contextlib import contextmanager

# Set this to True to have large FFTs raise a GalSimFFTSizeError  exception rather
# merely emit a GalSimFFTSizeWarning.
raise_fft_size_error = False

# Note to developers about which exception to throw.
#
# Aside from the below classes, which should be preferred for most errors, we also
# throw the following in some cases.
#
# TypeError:            Use this for errors that in a more strongly typed language would probably
#                       be a compiler error. For instance, it is used for the following errors:
#                       - a parameter with the wrong type
#                       - the wrong number of unnamed args when processing `*args` by hand.
#                       - missing or invalid kwargs when processing `**kwargs` by hand.
#
# OSError:              Use this for errors related to I/O, disk access, etc.
#
# NotImplementedError:  Use this for code that is not implemented by design and which will never
#                       be implemented. E.g. GSObject and Position use this for their __init__
#                       implementations, since it is invalid to instantiate the base class.
#                       Use GalSimNotImplementedError for features that might someday be
#                       implemented.
#
# AttributeError:       Use this only for an attempt to access an attribute that an object does not
#                       have. Like TypeError, this should be reserved for things that a more
#                       strongly typed language would catch at compile time. We don't currently
#                       raise this anywhere in GalSim.
#
# RuntimeError:         Don't use this. Use GalSimError (or a subclass) for any run-time errors.
#
# ValueError:           Don't use this. Use one of the below exceptions that derive from ValueError.
#
# KeyError:             Don't use this. Use GalSimKeyError instead
#
# IndexError:           Don't use this. Use GalSimIndexError instead.
#
# std::runtime_error:   Use this for errors in the C++ layer, and use the convert_cpp_errors()
#                       context to convert these errors into GalSimErrors. E.g. GSFitsWCS._invert_pv
#                       uses this for non-convergence, which gets converted into GalSimError in
#                       the Python layer.
#                       When possible, it is preferable to guard against any such events by making
#                       appropriate checks in the Python layer before dropping down into C++.
#                       E.g. Image checks for anything that might cause the C++ Image class to
#                       throw an exception and raises some kind of GalSim exception first.
#                       Nonetheless, it is good practice to use the `with convert_cpp_errors()`
#                       context for all calls to the C++ layer, just in case.
#
# GalSim-specific error classes:
# ------------------------------
#
# GalSimError:          Use this for what would normally be a RuntimeError. Usually some exceptional
#                       occurrence in otherwise correct code. E.g. an algorithm not converging or
#                       a singular matrix encountered. It can also be used when the program does
#                       things out of order; e.g. PowerSpectrum raises this when getShear and the
#                       like are called before `buildGrid`. This is also the catch-all exception
#                       to use when none of the other GalSim exceptions are appropriate.
#
# GalSimValueError:     Use this when a user provides an invalid value for a parameter.
#                       Note: it has an optional argument to give a list of allowed values when
#                       that is appropriate.
#
# GalSimKeyError        Use this for accessing a dict-like object with an invalid key. E.g.
#                       FitsHeader and Catalog raise this for accessing invalid columns.
#
# GalSimIndexError      Use this for the equivalent of accessing a list-like object with an
#                       invalid index. E.g. RealGalaxyCatalog and Catalog raise this for accessing
#                       invalid rows.
#
# GalSimRangeError:     Use this when a user provides an value outside of some allowed range.
#                       You should also give the min/max values of the allowed range. The max
#                       is optional, because it's not uncommon for there to be no upper limit.
#                       If only the upper limit is relevant and not the lower limit, you may
#                       use min=None to indicate this.
#
# GalSimBoundsError:    Use this when a position is outside its allowed bounds. It's basically
#                       the same as GalSimRangeError, but in two dimensions.
#
# GalSimUndefinedBoundsError:   Use this when the user tries to perform an operation on an
#                               Image with undefined bounds (and which requires the bounds to be
#                               defined).
#
# GalSimImmutableError: Use this when the user tries to modify an immutable Image in some way.
#
# GalSimIncompatibleValuesError:    Use this when two or more parameters are invalid when used
#                                   in combination. E.g. providing more than one size parameter
#                                   to Moffat, Sersic, Gaussian, etc. The conflicting values
#                                   should be given as extra keywords to the constructor, which
#                                   are mentioned in the error message.
#                                   Note: if one of the conflicting values is self (e.g. adding two
#                                   SEDs with different redshifts), then don't name the kwarg self.
#                                   Instead use something like `self_sed=self`.
#
# GalSimSEDError:       Use this when an SED is required to be either spectral or dimensionless,
#                       and the other kind of SED is provided.
#
# GalSimHSMError:       Use this for errors from the HSM algorithm.  They are emitted in C++, but
#                       we use `with convert_cpp_errors(GalSimHSMError):` to convert them.
#
# GalSimConfigError:    Use this for errors processing a config dict.
#
# GalSimConfigValueError:   Use this when a config dict has a value that is invalid. Basically,
#                           whenever you would normally use GalSimValueError when processing
#                           a config dict, you should use this instead.
#
# GalSimNotImplementedError:  Use this for features that we have not yet implemented, but which may
#                             be implemented someday. So it's not a necessarily invalid usage, just
#                             something that doesn't work currently.

class GalSimError(RuntimeError):
    """The base class for GalSim-specific run-time errors.
    """
    # Minimal version of these to make GalSimError reprable and picklable.
    def __repr__(self): return 'galsim.GalSimError(%r)'%(str(self))
    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __hash__(self): return hash(repr(self))


class GalSimValueError(GalSimError, ValueError):
    """A GalSim-specific exception class indicating that some user-input value is invalid.

    Attributes:
        value:          The invalid value
        allowed_values: A list of allowed values if appropriate (may be None)
    """
    def __init__(self, message, value, allowed_values=None):
        self.message = message
        self.value = value
        self.allowed_values = allowed_values

        message += " Value {0!s}".format(value)
        if allowed_values:
            message += " not in {0!s}".format(allowed_values)
        super(GalSimValueError, self).__init__(message)

    def __repr__(self):
        return 'galsim.GalSimValueError(%r,%r,%r)'%(self.message, self.value, self.allowed_values)
    def __reduce__(self):  # Need to override this whenever constructor take extra params
        return GalSimValueError, (self.message, self.value, self.allowed_values)


class GalSimKeyError(GalSimError, KeyError):
    """A GalSim-specific exception class indicating an attempt to access a dict-like object
    with an invalid key.

    Attributes:
        key:        The invalid key
    """
    def __init__(self, message, key):
        self.message = message
        self.key = key
        super(GalSimKeyError, self).__init__(message, key)  # Need to pass key or pickle fails.

    def __str__(self):
        return self.message + " Key {0!s}".format(self.key)

    def __repr__(self):
        return 'galsim.GalSimKeyError(%r,%r)'%(self.message, self.key)


class GalSimIndexError(GalSimError, IndexError):
    """A GalSim-specific exception class indicating an attempt to access a list-like object
    with an invalid index.

    Attributes:
        index:      The invalid index
    """
    def __init__(self, message, index):
        self.message = message
        self.index = index
        super(GalSimIndexError, self).__init__(message, index)

    def __str__(self):
        return self.message + " Index {0!s}".format(self.index)

    def __repr__(self):
        return 'galsim.GalSimIndexError(%r,%r)'%(self.message, self.index)


class GalSimRangeError(GalSimError, ValueError):
    """A GalSim-specific exception class indicating that some user-input value is
    outside of the allowed range of values.

    Attributes:
        value:      The invalid value
        min:        The minimum allowed value (may be None)
        max:        The maximum allowed value (may be None)
    """
    def __init__(self, message, value, min, max=None):
        self.message = message
        self.value = value
        self.min = min
        self.max = max

        message += " Value {0!s} not in range [{1!s}, {2!s}]".format(value, min, max)
        super(GalSimRangeError, self).__init__(message)

    def __repr__(self):
        return 'galsim.GalSimRangeError(%r,%r,%r,%r)'%(self.message, self.value, self.min, self.max)
    def __reduce__(self):
        return GalSimRangeError, (self.message, self.value, self.min, self.max)


class GalSimBoundsError(GalSimError, ValueError):
    """A GalSim-specific exception class indicating that some user-input position is
    outside of the allowed bounds.

    Attributes:
        pos:        The invalid position
        bounds:     The bounds in which it was expected to fall
    """
    def __init__(self, message, pos, bounds):
        self.message = message
        self.pos = pos
        self.bounds = bounds

        message += " {0!s} not in {1!s}".format(pos, bounds)
        super(GalSimBoundsError, self).__init__(message)

    def __repr__(self):
        return 'galsim.GalSimBoundsError(%r,%r,%r)'%(self.message, self.pos, self.bounds)
    def __reduce__(self):
        return GalSimBoundsError, (self.message, self.pos, self.bounds)


class GalSimUndefinedBoundsError(GalSimError):
    """A GalSim-specific exception class indicating an attempt to access the extent of
    a `Bounds` instance that has not yet been defined.
    """
    def __repr__(self):
        return 'galsim.GalSimUndefinedBoundsError(%r)'%(str(self))


class GalSimImmutableError(GalSimError):
    """A GalSim-specific exception class indicating an attempt to modify an immutable image.

    Attributes:
        image:      The image that the user attempted to modify
    """
    def __init__(self, message, image):
        self.message = message
        self.image = image

        message += " Image: {0!s}".format(image)
        super(GalSimImmutableError, self).__init__(message)

    def __repr__(self):
        return 'galsim.GalSimImmutableError(%r,%r)'%(self.message, self.image)
    def __reduce__(self):
        return GalSimImmutableError, (self.message, self.image)


class GalSimIncompatibleValuesError(GalSimError, ValueError, TypeError):
    """A GalSim-specific exception class indicating that 2 or more user-input values are
    incompatible as given.

    Attributes:
        values:     A dict of {name : value} giving the values that in combination are invalid.
    """
    def __init__(self, message, values={}, **kwargs):
        self.message = message
        self.values = dict(values, **kwargs)

        message += " Values {0!s}".format(self.values)
        super(GalSimIncompatibleValuesError, self).__init__(message)

    # Note: the repr of values can rearrange the items, but the dicts should compare equal.
    def __eq__(self, other):
        return (self is other or
                (isinstance(other, GalSimIncompatibleValuesError) and
                 self.message == other.message and
                 self.values == other.values))
    def __repr__(self):
        return 'galsim.GalSimIncompatibleValuesError(%r,%r)'%(self.message, self.values)
    def __reduce__(self):
        return GalSimIncompatibleValuesError, (self.message, self.values)


class GalSimSEDError(GalSimError, TypeError):
    """A GalSim-specific exception class indicating an attempt to do something invalid for the
    kind of `SED` that is present.  Typically involving a dimensionless `SED` where a spectral
    `SED` is required (or vice versa).

    Attributes:
        sed:        The invalid `SED`
    """
    def __init__(self, message, sed):
        self.message = message
        self.sed = sed

        message += " SED: {0!s}".format(sed)
        super(GalSimSEDError, self).__init__(message)

    def __repr__(self):
        return 'galsim.GalSimSEDError(%r,%r)'%(self.message, self.sed)
    def __reduce__(self):
        return GalSimSEDError, (self.message, self.sed)


class GalSimHSMError(GalSimError):
    """A GalSim-specific exception class indicating some kind of failure of the HSM algorithms
    """
    def __repr__(self):
        return 'galsim.GalSimHSMError(%r)'%(str(self))


class GalSimFFTSizeError(GalSimError):
    """A GalSim-specific exception class indicating that a requested FFT exceeds the relevant
    maximum_fft_size.

    Attributes:
        size:       The size that was deemed too large
        mem:        The estimated memory that would be required (in GB) for the FFT.
    """
    def __init__(self, message, size):
        self.message = message
        self.size = size
        self.mem = size * size * 24. / 1024**3
        message += "\nThe required FFT size would be {0} x {0}, which requires ".format(size)
        message += "{0:.2f} GB of memory.\n".format(self.mem)
        message += "If you can handle the large FFT, you may update gsparams.maximum_fft_size."
        super(GalSimFFTSizeError, self).__init__(message)

    def __repr__(self):
        return 'galsim.GalSimFFTSizeError(%r,%r)'%(self.message, self.size)
    def __reduce__(self):
        return GalSimFFTSizeError, (self.message, self.size)


class GalSimConfigError(GalSimError, ValueError):
    """A GalSim-specific exception class indicating some kind of failure processing a
    configuration file.
    """
    def __repr__(self):
        return 'galsim.GalSimConfigError(%r)'%(str(self))


class GalSimConfigValueError(GalSimValueError, GalSimConfigError):
    """A GalSim-specific exception class indicating that a config entry has an invalid value.

    Attributes:
        value:          The invalid value
        allowed_values: A list of allowed values if appropriate (may be None)
    """
    def __repr__(self):
        return 'galsim.GalSimConfigValueError(%r,%r,%r)'%(
            self.message, self.value, self.allowed_values)
    def __reduce__(self):
        return GalSimConfigValueError, (self.message, self.value, self.allowed_values)


class GalSimNotImplementedError(GalSimError, NotImplementedError):
    """A GalSim-specific exception class indicating that the feature being attempted is not
    currently implemented.

    If this is a feature you feel you need, please open an issue about it at

        https://github.com/GalSim-developers/GalSim/issues

    Even better, feel free to offer to contribute code to implement the feature.
    """
    def __repr__(self):
        return 'galsim.GalSimNotImplementedError(%r)'%(str(self))


# Note: Can use galsim_warn to raise warnings with this warning class.
class GalSimWarning(UserWarning):
    """The base class for GalSim-emitted warnings.
    """
    def __repr__(self): return 'galsim.GalSimWarning(%r)'%(str(self))
    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __hash__(self): return hash(repr(self))


# Note: By default python ignores DeprecationWarnings.  Apparently they are really
#       for python system deprecations.  GalSim deprecations are thus only subclassed from
#       GalSimWarning, not DeprecationWarning.
class GalSimDeprecationWarning(GalSimWarning):
    """A GalSim-specific warning class used for deprecation warnings.
    """
    def __repr__(self): return 'galsim.GalSimDeprecationWarning(%r)'%(str(self))

class GalSimFFTSizeWarning(GalSimWarning):
    """A GalSim-specific warning class indicating that a requested FFT exceeds the relevant
    maximum_fft_size.

    Attributes:
        size:       The size that was deemed too large
        mem:        The estimated memory that would be required (in GB) for the FFT.
    """
    def __init__(self, message, size):
        self.message = message
        self.size = size
        self.mem = size * size * 24. / 1024**3
        message += "\nThe required FFT size would be {0} x {0}, which requires ".format(size)
        message += "{0:.2f} GB of memory.\n".format(self.mem)
        message += "If you can handle the large FFT and want to suppress this warning,\n"
        message += "you may update gsparams.maximum_fft_size."
        super(GalSimFFTSizeWarning, self).__init__(message)
    def __repr__(self):
        return 'galsim.GalSimFFTSizeWarning(%r,%r)'%(self.message, self.size)
    def __reduce__(self):
        return GalSimFFTSizeWarning, (self.message, self.size)

def galsim_warn_fft(message, size):
    if raise_fft_size_error:
        raise GalSimFFTSizeError(message, size)
    else:
        warnings.warn(GalSimFFTSizeWarning(message, size))

@contextmanager
def convert_cpp_errors(error_type=GalSimError):
    try:
        yield
    except RuntimeError as err:
        raise error_type(str(err)) from None

def galsim_warn(message):
    """A helper function for emitting a GalSimWarning with the given message
    """
    warnings.warn(message, GalSimWarning)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from astropy.units import Quantity, Unit

from .util import LoggerWrapper
from .value import ParseValue, GetAllParams, GetIndex
from .input import RegisterInputConnectedType
from .bandpass import BuildBandpass
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..sed import SED
from ..utilities import basestring, LRU_Cache

# This module-level dict will store all the registered SED types.
# See the RegisterSEDType function at the end of this file.
# The keys are the (string) names of the SED types, and the values will be builders that know
# how to build the SED object.
valid_sed_types = {}

def BuildSED(config, key, base, logger=None):
    """Read the SED parameters from config[key] and return a constructed SED object.

    Parameters:
        config:     A dict with the configuration information.
        key:        The key name in config indicating which object to build.
        base:       The base dict of the configuration.
        logger:     Optionally, provide a logger for logging debug statements. [default: None]

    Returns:
        (sed, safe) where sed is an SED instance, and safe is whether it is safe to reuse.
    """
    logger = LoggerWrapper(logger)
    logger.debug('obj %d: Start BuildSED key = %s',base.get('obj_num',0),key)

    param = config[key]

    # Check for direct value, else get the SED type
    if isinstance(param, SED):
        return param, True
    elif isinstance(param, basestring) and (param[0] == '$' or param[0] == '@'):
        return ParseValue(config, key, base, None)
    elif isinstance(param, dict):
        sed_type = param.get('type','FileSED')
    else:
        raise GalSimConfigError("%s must be either an SED or a dict"%key)

    # For these two, just do the usual ParseValue function.
    if sed_type in ('Eval', 'Current'):
        return ParseValue(config, key, base, None)

    # Check if we can use the current cached object
    index, index_key = GetIndex(param, base)
    if 'current' in param:
        csed, csafe, cvalue_type, cindex, cindex_key = param['current']
        if cindex == index:
            logger.debug('obj %d: The SED object is already current', base.get('obj_num',0))
            logger.debug('obj %d: index_key = %s, index = %d',base.get('obj_num',0),
                         cindex_key, cindex)
            return csed, csafe

    if sed_type not in valid_sed_types:
        raise GalSimConfigValueError("Invalid sed.type.", sed_type, list(valid_sed_types.keys()))
    logger.debug('obj %d: Building sed type %s', base.get('obj_num',0), sed_type)
    builder = valid_sed_types[sed_type]
    sed, safe = builder.buildSED(param, base, logger)
    logger.debug('obj %d: sed = %s', base.get('obj_num',0), sed)

    param['current'] = sed, safe, SED, index, index_key

    return sed, safe


class SEDBuilder:
    """A base class for building SED objects.

    The base class defines the call signatures of the methods that any derived class should follow.
    """
    def buildSED(self, config, base, logger):
        """Build the SED based on the specifications in the config dict.

        Note: Sub-classes must override this function with a real implementation.

        Parameters:
            config:     The configuration dict for the SED type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed SED object.
        """
        raise NotImplementedError("The %s class has not overridden buildSED"%self.__class__)


def _read_sed_file(file_name, wave_type, flux_type):
    return SED(file_name, wave_type, flux_type)
read_sed_file = LRU_Cache(_read_sed_file)

class FileSEDBuilder(SEDBuilder):
    """A class for loading an SED from a file

    FileSED expected the following parameters:

        file_name (required)    The file to load
        wave_type(required)     The units (nm or Ang) of the wavelengths in the file
        flux_type (required)    Which kind of flux values are in the file
                                Allowed values: flambda, fnu, fphotons, 1
    """
    def buildSED(self, config, base, logger):
        """Build the SED based on the specifications in the config dict.

        Parameters:
            config:     The configuration dict for the SED type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed SED object.
        """
        logger = LoggerWrapper(logger)

        req = {
            'file_name': str,
            'wave_type': (Unit, str),
            'flux_type': (Unit, str),
        }
        opt = {
            'norm_flux_density': (float, Quantity),
            'norm_wavelength': (float, Quantity),
            'norm_flux': float,
            'redshift': float
        }
        ignore = ['norm_bandpass']

        kwargs, safe = GetAllParams(config, base, req=req, opt=opt, ignore=ignore)

        file_name = kwargs.pop('file_name')

        norm_flux_density = kwargs.pop('norm_flux_density', None)
        norm_wavelength = kwargs.pop('norm_wavelength', None)
        norm_flux = kwargs.pop('norm_flux', None)
        redshift = kwargs.pop('redshift', 0.)
        wave_type = kwargs.pop('wave_type')
        flux_type = kwargs.pop('flux_type')

        logger.info("Using SED file: %s",file_name)
        sed = read_sed_file(file_name, wave_type, flux_type)
        if norm_flux_density is not None:
            sed = sed.withFluxDensity(norm_flux_density, wavelength=norm_wavelength)
        elif norm_flux:
            bandpass, safe1 = BuildBandpass(config, 'norm_bandpass', base, logger)
            sed = sed.withFlux(norm_flux, bandpass=bandpass)
            safe = safe and safe1
        sed = sed.atRedshift(redshift)

        return sed, safe

def RegisterSEDType(sed_type, builder, input_type=None):
    """Register a SED type for use by the config apparatus.

    Parameters:
        sed_type:       The name of the type in the config dict.
        builder:        A builder object to use for building the SED object.  It should
                        be an instance of a subclass of SEDBuilder.
        input_type:     If the SED builder utilises an input object, give the key name of the
                        input type here.  (If it uses more than one, this may be a list.)
                        [default: None]
    """
    valid_sed_types[sed_type] = builder
    RegisterInputConnectedType(input_type, sed_type)

RegisterSEDType('FileSED', FileSEDBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os

import galsim
from galsim_test_helpers import *


# Get whatever version of pyfits or astropy we are using
from galsim._pyfits import pyfits


@timer
def test_read():
    """Test reading a FitsHeader from an existing FITS file
    """
    tpv_len = 215

    def check_tpv(header):
        """Check that the header object has correct values from the tpv.fits file
        """
        # Check using a few different access methods.
        assert header['TIME-OBS'] == '04:28:14.105'
        assert header.get('FILTER') == 'I'
        assert header['AIRMASS'] == 1.185
        assert len(header) == tpv_len
        assert 'ADC' in header
        assert ('FILPOS',6) in header.items()
        assert ('FILPOS',6) in header.iteritems()
        assert 'OBSERVAT' in header.keys()
        assert 'OBSERVAT' in header.iterkeys()
        assert 54384.18627436 in header.values() # MJD-OBS
        assert 54384.18627436 in header.itervalues()

    file_name = 'tpv.fits'
    dir = 'fits_files'
    # First option: give a file_name
    header = galsim.FitsHeader(file_name=os.path.join(dir,file_name))
    check_tpv(header)
    check_pickle(header)
    # Let the FitsHeader init handle the dir
    header = galsim.FitsHeader(file_name=file_name, dir=dir)
    check_tpv(header)
    check_pickle(header)
    # If the first arg is a str, then it should be interpreted as a file name
    header = galsim.FitsHeader(file_name, dir=dir)
    check_tpv(header)
    # If you pass in a pyfits hdulist, that should also work
    with pyfits.open(os.path.join(dir,file_name)) as hdu_list:
        header = galsim.FitsHeader(hdu_list=hdu_list)
    check_tpv(header)
    check_pickle(header)
    # Can explicitly give an hdu number to use.  In this case, there is only 1, so need to use 0.
    with pyfits.open(os.path.join(dir,file_name)) as hdu_list:
        header = galsim.FitsHeader(hdu_list=hdu_list, hdu=0)
    check_tpv(header)
    check_pickle(header)
    # Can explicitly give an hdu number to use.  In this case, there is only 1, so need to use 0.
    header = galsim.FitsHeader(file_name=file_name, dir=dir, hdu=0)
    check_tpv(header)
    check_pickle(header)
    # Can also pass the hdu itself.
    with pyfits.open(os.path.join(dir,file_name)) as hdu_list:
        header = galsim.FitsHeader(hdu_list=hdu_list[0])
    check_tpv(header)
    check_pickle(header)
    # If you pass in a pyfits Header object, that should also work
    with pyfits.open(os.path.join(dir,file_name)) as hdu_list:
        header = galsim.FitsHeader(header=hdu_list[0].header)
    check_tpv(header)
    check_pickle(header)
    # The header is the first parameter, so don't need to name it.
    with pyfits.open(os.path.join(dir,file_name)) as hdu_list:
        header = galsim.FitsHeader(hdu_list[0].header)
    check_tpv(header)
    # FitsHeader can read from a compressed file too
    header = galsim.FitsHeader(file_name=file_name + '.gz', dir=dir, compression='auto')
    check_tpv(header)
    check_pickle(header)
    header = galsim.FitsHeader(file_name=file_name + '.gz', dir=dir, compression='gzip')
    check_tpv(header)
    check_pickle(header)

    assert_raises(TypeError, galsim.FitsHeader, file_name=file_name, header=header)
    with pyfits.open(os.path.join(dir,file_name)) as hdu_list:
        assert_raises(TypeError, galsim.FitsHeader, file_name=file_name, hdu_list=hdu_list)
        assert_raises(TypeError, galsim.FitsHeader, header=header, hdu_list=hdu_list)

    # Remove an item from the header
    # Start with file_name constructor, to test that the repr is changed by the edit.
    orig_header = header
    header = galsim.FitsHeader(file_name=os.path.join(dir,file_name))
    assert header == orig_header
    del header['AIRMASS']
    assert 'AIRMASS' not in header
    assert len(header) == tpv_len-1
    assert header != orig_header
    check_pickle(header)

    # Should be able to get with a default value if the key is not present
    assert header.get('AIRMASS', 2.0) == 2.0
    # key should still not be in the header
    assert 'AIRMASS' not in header
    assert len(header) == tpv_len-1
    assert header != orig_header

    # Add items to a header
    header['AIRMASS'] = 2
    assert header.get('AIRMASS') == 2
    assert header != orig_header

    # Pop does a similar thing:
    assert header.pop('AIRMASS') == 2.0
    assert 'AIRMASS' not in header

    # Works if not preset, given default
    assert header.pop('AIRMASS', 2.0) == 2.0
    assert 'AIRMASS' not in header
    header['AIRMASS'] = 2
    assert header['AIRMASS'] == 2

    # Get real value if preset and given default value
    assert header.pop('AIRMASS', 1.9) == 2.0
    assert 'AIRMASS' not in header
    header['AIRMASS'] = 2
    assert header['AIRMASS'] == 2

    # Overwrite an existing value
    header['AIRMASS'] = 1.7
    assert header.get('AIRMASS') == 1.7
    assert header != orig_header

    # Set with a comment field
    header['AIRMASS'] = (1.9, 'The airmass of the observation')
    assert header.get('AIRMASS') == 1.9
    assert header != orig_header

    # Update with a dict
    d = { 'AIRMASS' : 1.185 }
    header.update(d)
    assert header.get('AIRMASS') == 1.185

    # Extend is similar, but skip duplicates, and must be a FitsHeader
    h = galsim.FitsHeader({ 'AIRMASS' : 1.2, 'RA': 123 })
    header.extend(h)
    assert header['AIRMASS'] == 1.185  # unchanged
    assert header['RA'] == 123

    # If replace = true, then update to the new values.
    h['DEC'] = -22.5
    header.extend(h, replace=True)
    assert header['AIRMASS'] == 1.2  # changed
    assert header['RA'] == 123
    assert header['DEC'] == -22.5

    # We are essentially back to where we started, except the len won't be right.
    # Deleting a key removed an item each time, but setting it overwrote a blank item.
    # But if we add back another few of these, we should be back to the original values.
    header.append('','', useblanks=False)
    header.append('','', useblanks=False)
    header.append('','', useblanks=False)
    header['AIRMASS'] = 1.185
    check_tpv(header)
    check_pickle(header)
    assert header != orig_header  # It's still not equal, because the AIRMASS item is in a
                                  # different location in the list, which is relevant for equality.

    # Clear all values
    header.clear()
    assert 'AIRMASS' not in header
    assert 'FILTER' not in header
    assert len(header) == 0
    check_pickle(header)
    assert header != orig_header


@timer
def test_scamp():
    """Test that we can read in a SCamp .head file correctly
    """
    dir = 'fits_files'
    file_name = 'scamp.head'

    header = galsim.FitsHeader(file_name=file_name, dir=dir, text_file=True)
    # Just check a few values.  The full test of this file as a wcs is in test_wcs.py
    assert header['RADECSYS'] == 'FK5'
    assert header['MAGZEROP'] == 30.
    assert header['ASTINST'] == 39
    check_pickle(header)


def check_dict(d):
    def check_dict(header):
        """Check that the header object has correct values from the given dict
        """
        assert header['TIME-OBS'] == '04:28:14.105'
        assert header['FILTER'] == 'I'
        assert header['AIRMASS'] == 1.185
        assert len(header) == 3

    # Construct from a given dict
    header = galsim.FitsHeader(header = d)
    check_dict(header)
    check_pickle(header)

    # Start with a blank dict and add elements individually
    header = galsim.FitsHeader(header = {})
    check_pickle(header)
    for k in d:
        header[k] = d[k]
    check_dict(header)
    check_pickle(header)

    # Set with a comment field
    header = galsim.FitsHeader(header = {})
    for k in d:
        header[k] = (d[k], 'The value of ' + k)
    check_dict(header)
    check_pickle(header)

    # Use update
    header = galsim.FitsHeader({})
    header.update(d)
    check_dict(header)
    check_pickle(header)

    # Use default constructor
    header = galsim.FitsHeader()
    check_pickle(header)
    assert len(header) == 0
    header.update(d)
    check_dict(header)
    check_pickle(header)

@timer
def test_dict():
    """Test that we can create a FitsHeader from a dict
    """
    d = { 'TIME-OBS' : '04:28:14.105' ,
          'FILTER'   : 'I',
          'AIRMASS'  : 1.185 }
    check_dict(d)

@timer
def test_lowercase():
    """Test that lowercase keys are turned into uppercase.
    """
    d = { 'Time-Obs' : '04:28:14.105' ,
          'filter'   : 'I',
          'AirMAsS'  : 1.185 }
    check_dict(d)

@timer
def test_comments():
    """Test that comments are correctly added to the cards.
    """
    # cf. Issue #877
    # This test is taken directly from Bryan's failing example

    im1 = galsim.Image(1,1,dtype=float)
    im1.header = galsim.FitsHeader()
    im1.header["CARD"] = ("value","comment")
    galsim.fits.write(im1,os.path.join('output','test_comments_im1.fits'))

    im2 = galsim.Image(1,1,dtype=float)
    im2.header = {}
    im2.header["CARD"] = ("value","comment")
    galsim.fits.write(im1,os.path.join('output','test_comments_im2.fits'))

    im3 = galsim.Image(1,1,dtype=float)
    im3.header = {"CARD": ("value","comment")}
    galsim.fits.write(im1,os.path.join('output','test_comments_im3.fits'))

    im1 = galsim.fits.read(os.path.join('output','test_comments_im1.fits'), read_header=True)
    im2 = galsim.fits.read(os.path.join('output','test_comments_im2.fits'), read_header=True)
    im3 = galsim.fits.read(os.path.join('output','test_comments_im3.fits'), read_header=True)

    for im in [im1, im2, im3]:
        assert im.header['CARD'] == 'value'
        assert im.header.comment('CARD') == 'comment'


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.

import numpy as np
import sys
import warnings

from .image import Image
from .errors import GalSimRangeError, GalSimValueError, GalSimIncompatibleValuesError, galsim_warn

def applyNonlinearity(self, NLfunc, *args):
    """
    Applies the given non-linearity function (``NLfunc``) on the `Image` instance directly.

    This routine can transform the image in a non-linear manner specified by the user. However,
    the typical kind of non-linearity one sees in astronomical images is voltage non-linearity,
    also sometimes known as 'classical non-linearity', refers to the non-linearity in
    charge-to-voltage conversion process. This arises as charge gets integrated at the junction
    capacitance of the pixel node. Voltage non-linearity decreases signals at higher signal
    levels, causing the attenuation of brighter pixels. The image should include both the
    signal from the astronomical objects as well as the background level. Other detectors effects
    such as dark current and persistence (not currently included in GalSim) would also occur
    before the inclusion of nonlinearity.

    The argument ``NLfunc`` is a callable function (for example a lambda function, a
    `galsim.LookupTable`, or a user-defined function), possibly with arguments that need to be given
    as subsequent arguments to the ``applyNonlinearity`` function (after the ``NLfunc`` argument).
    ``NLfunc`` should be able to take a 2d NumPy array as input, and return a NumPy array of the
    same shape.  It should be defined such that it outputs the final image with nonlinearity
    included (i.e., in the limit that there is no nonlinearity, the function should return the
    original image, NOT zero). The image should be in units of electrons when this routine is being
    used to generate classical non-linearity. When used for other purposes, the units can be in
    electrons or in ADU, as found appropriate by the user.

    Examples::

        >>> f = lambda x: x + (1.e-7)*(x**2)
        >>> img.applyNonlinearity(f)

        >>> f = lambda x, beta1, beta2: x - beta1*x*x + beta2*x*x*x
        >>> img.applyNonlinearity(f, 1.e-7, 1.e-10)

    On calling the method, the `Image` instance ``img`` is transformed by the user-defined function
    ``f`` with ``beta1`` = 1.e-7 and ``beta2`` = 1.e-10.

    Parameters:
        NLfunc:  The function that maps the input image pixel values to the output image pixel
                 values.
        *args:   Any subsequent arguments are passed along to the NLfunc function.

    """

    # Extract out the array from Image since not all functions can act directly on Images
    result = NLfunc(self.array,*args)
    if not isinstance(result, np.ndarray):
        raise GalSimValueError("NLfunc does not return a NumPy array.", NLfunc)
    if self.array.shape != result.shape:
        raise GalSimValueError("NLfunc does not return a NumPy array of the same shape as input.",
                               NLfunc)
    self.array[:,:] = result


def addReciprocityFailure(self, exp_time, alpha, base_flux):
    r"""
    Accounts for the reciprocity failure and includes it in the original `Image` directly.

    Reciprocity, in the context of photography, is the inverse relationship between the incident
    flux (I) of a source object and the exposure time (t) required to produce a given response (p)
    in the detector, i.e., p = I*t. At very low (also at high) levels of incident flux, deviation
    from this relation is observed, leading to reduced sensitivity at low flux levels. The pixel
    response to a high flux is larger than its response to a low flux. This flux-dependent non-
    linearity is known as 'Reciprocity Failure' and is known to happen in photographic films since
    1893. Interested users can refer to http://en.wikipedia.org/wiki/Reciprocity_(photography)

    CCDs are not known to suffer from this effect. HgCdTe detectors that are used for near infrared
    astrometry, although to an extent much lesser than the photographic films, are found to
    exhibit reciprocity failure at low flux levels. The exact mechanism of this effect is unknown
    and hence we lack a good theoretical model. Many models that fit the empirical data exist and
    a common relation is

    .. math::
            \frac{p_R}{p} = \left(1 + \alpha \log_{10}\left(\frac{p}{t}\right)
                               - \alpha \log_{10}\left(\frac{p^\prime}{t^\prime}\right)\right)

    where :math:`t` is the exposure time (in units of seconds), :math:`p` is the pixel response
    (in units of electrons) and :math:`p_R` is the response if the reciprocity relation fails to
    hold. :math:`p^\prime/t^\prime` is the count rate (in electrons/second) corresponding to the
    photon flux (base flux) at which the detector is calibrated to have its nominal gain.
    \alpha is the parameter in the model, measured in units of per decade and varies with detectors
    and the operating temperature. The functional form for the reciprocity failure is motivated
    empirically from the tests carried out on H2RG detectors.

    See for reference Fig. 1 and Fig. 2 of http://arxiv.org/abs/1106.1090. Since :math:`p_R/p`
    remains close to unity over a wide range of flux, we convert this relation to a power law by
    approximating :math:`p_R/p \approx 1 + \log(p_R/p)`. This gives a relation that is better
    behaved than the logarithmic relation at low flux levels.

    .. math::
            \frac{p_R}{p} = \left(\frac{p/t}{p^\prime/t^\prime}\right)^\frac{\alpha}{\log(10)}.

    Because of how this function is defined, the input image must have non-negative pixel
    values for the resulting image to be well-defined. Negative pixel values result in 'nan's.
    The image should be in units of electrons, or if it is in ADU, then the value passed to
    exp_time should be the exposure time divided by the nominal gain. The image should include
    both the signal from the astronomical objects as well as the background level.  The addition of
    nonlinearity should occur after including the effect of reciprocity failure.

    Parameters:
        exp_time:   The exposure time (t) in seconds, which goes into the expression for
                    reciprocity failure given in the docstring.
        alpha:      The alpha parameter in the expression for reciprocity failure, in
                    units of 'per decade'.
        base_flux:  The flux (:math:`p^\prime/t^\prime`) at which the gain is calibrated to have
                    its nominal value.
    """

    if alpha < 0.:
        raise GalSimRangeError("Invalid value of alpha, must be >= 0",
                               alpha, 0, None)
    if exp_time < 0.:
        raise GalSimRangeError("Invalid value of exp_time, must be >= 0",
                               exp_time, 0, None)
    if base_flux < 0.:
        raise GalSimRangeError("Invalid value of base_flux, must be >= 0",
                               base_flux, 0, None)

    if np.any(self.array<0):
        galsim_warn("One or more pixel values are negative and will be set as 'nan'.")

    p0 = exp_time*base_flux
    a = alpha/np.log(10)
    with warnings.catch_warnings():
        # If self.array has negative values, then this might raise a RuntimeWarning.
        # We already gave a more useful warning above, so ignore it here.
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        self.applyNonlinearity(lambda x,x0,a: (x**(a+1))/(x0**a), p0, a)

def applyIPC(self, IPC_kernel, edge_treatment='extend', fill_value=None, kernel_nonnegativity=True,
    kernel_normalization=True):
    """
    Applies the effect of interpixel capacitance to the `Image` instance.

    In NIR detectors, the quantity that is sensed is not the charge as in CCDs, but a voltage that
    relates to the charge present within each pixel. The voltage read at a given pixel location is
    influenced by the charges present in the neighboring pixel locations due to capacitive
    coupling of sense nodes.

    This interpixel capacitance is approximated as a linear effect that can be described by a 3x3
    kernel that is convolved with the image. The kernel must be an `Image` instance and could be
    intrinsically anisotropic. A sensible kernel must have non-negative entries and must be
    normalized such that the sum of the elements is 1, in order to conserve the total charge.
    The (1,1) element of the kernel is the contribution to the voltage read at a pixel from the
    electrons in the pixel to its bottom-left, the (1,2) element of the kernel is the contribution
    from the charges to its left and so on.

    The argument 'edge_treatment' specifies how the edges of the image should be treated, which
    could be in one of the three ways:

    1. 'extend': The kernel is convolved with the zero-padded image, leading to a larger
       intermediate image. The central portion of this image is returned.  [default]
    2. 'crop': The kernel is convolved with the image, with the kernel inside the image completely.
       Pixels at the edges, where the center of the kernel could not be placed, are set to the
       value specified by 'fill_value'. If 'fill_value' is not specified or set to 'None', then
       the pixel values in the original image are retained. The user can make the edges invalid
       by setting fill_value to numpy.nan.
    3. 'wrap': The kernel is convolved with the image, assuming periodic boundary conditions.

    The size of the image array remains unchanged in all three cases.

    Parameters:
        IPC_kernel:            A 3x3 `Image` instance that is convolved with the `Image` instance
        edge_treatment:        Specifies the method of handling edges and should be one of
                               'crop', 'extend' or 'wrap'. See above for details.
                               [default: 'extend']
        fill_value:            Specifies the value (including nan) to fill the edges with when
                               edge_treatment is 'crop'. If unspecified or set to 'None', the
                               original pixel values are retained at the edges. If
                               edge_treatment is not 'crop', then this is ignored.
        kernel_nonnegativity:  Specify whether the kernel should have only non-negative
                               entries.  [default: True]
        kernel_normalization:  Specify whether to check and enforce correct normalization for
                               the kernel.  [default: True]
    """

    # IPC kernel has to be a 3x3 Image
    ipc_kernel = IPC_kernel.array
    if not ipc_kernel.shape==(3,3):
        raise GalSimValueError("IPC kernel must be an Image instance of size 3x3.", IPC_kernel)

    # Check for non-negativity of the kernel
    if kernel_nonnegativity and (ipc_kernel<0).any():
        raise GalSimValueError("IPC kernel must not contain negative entries", IPC_kernel)

    # Check and enforce correct normalization for the kernel
    if kernel_normalization and abs(ipc_kernel.sum()-1) > 10.*np.finfo(ipc_kernel.dtype.type).eps:
        galsim_warn("The entries in the IPC kernel did not sum to 1. Scaling the kernel to "
                    "ensure correct normalization.")
        IPC_kernel = IPC_kernel/ipc_kernel.sum()

    # edge_treatment can be 'extend', 'wrap' or 'crop'
    if edge_treatment=='crop':
        # Simply re-label the array of the Image instance
        pad_array = self.array
    elif edge_treatment=='extend':
        # Copy the array of the Image instance and pad with zeros
        pad_array = np.zeros((self.array.shape[0]+2,self.array.shape[1]+2))
        pad_array[1:-1,1:-1] = self.array
    elif edge_treatment=='wrap':
        # Copy the array of the Image instance and pad with zeros initially
        pad_array = np.zeros((self.array.shape[0]+2,self.array.shape[1]+2))
        pad_array[1:-1,1:-1] = self.array
        # and wrap around the edges
        pad_array[0,:] = pad_array[-2,:]
        pad_array[-1,:] = pad_array[1,:]
        pad_array[:,0] = pad_array[:,-2]
        pad_array[:,-1] = pad_array[:,1]
    else:
        raise GalSimValueError("Invalid edge_treatment.", edge_treatment,
                               ('extend', 'wrap', 'crop'))

    # Generating different segments of the padded array
    center = pad_array[1:-1,1:-1]
    top = pad_array[2:,1:-1]
    bottom = pad_array[:-2,1:-1]
    left = pad_array[1:-1,:-2]
    right = pad_array[1:-1,2:]
    topleft = pad_array[2:,:-2]
    bottomright = pad_array[:-2,2:]
    topright = pad_array[2:,2:]
    bottomleft = pad_array[:-2,:-2]

    # Ensure that the origin is (1,1)
    kernel = IPC_kernel.view()
    kernel.setOrigin(1,1)

    # Generating the output array, with 2 rows and 2 columns lesser than the padded array
    # Image values have been used to make the code look more intuitive
    out_array = kernel(1,3)*topleft + kernel(2,3)*top + kernel(3,3)*topright + \
        kernel(1,2)*left + kernel(2,2)*center + kernel(3,2)*right + \
        kernel(1,1)*bottomleft + kernel(2,1)*bottom + kernel(3,1)*bottomright

    if edge_treatment=='crop':
        self.array[1:-1,1:-1] = out_array
        #Explicit edge effects handling with filling the edges with the value given in fill_value
        if fill_value is not None:
            self.array[0,:] = fill_value
            self.array[-1,:] = fill_value
            self.array[:,0] = fill_value
            self.array[:,-1] = fill_value
    else:
        self.array[:,:] = out_array

def applyPersistence(self,imgs,coeffs):
    """
    Applies the effects of persistence to the `Image` instance.

    Persistence refers to the retention of a small fraction of the signal after resetting the
    imager pixel elements. The persistence signal of a previous exposure is left in the pixel even
    after several detector resets. This effect is most likely due to charge traps in the material.
    Laboratory tests on the Roman Space Telescope CMOS detectors show that if exposures and
    readouts are taken in a fixed cadence, the persistence signal can be given as a linear
    combination of prior pixel values that can be added to the current image.

    This routine takes in a list of `Image` instances and adds them to `Image` weighted by the
    values passed on to 'coeffs'. The pixel values of the `Image` instances in the list must
    correspond to the electron counts before the readout. This routine does NOT keep track of
    realistic dither patterns. During the image simulation process, the user has to queue a list of
    previous `Image` instances (imgs) outside the routine by inserting the latest image in the
    beginning of the list and deleting the oldest image. The values in 'coeffs' tell how much of
    each `Image` is to be added. This usually remains constant in the image generation process.

    Parameters:
        imgs:       A list of previous `Image` instances that still persist.
        coeffs:     A list of floats that specifies the retention factors for the corresponding
                    `Image` instances listed in 'imgs'.
    """
    if not len(imgs)==len(coeffs):
        raise GalSimIncompatibleValuesError("The length of 'imgs' and 'coeffs' must be the same",
                                            imgs=imgs, coeffs=coeffs)
    for img,coeff in zip(imgs,coeffs):
        self += coeff*img

def quantize(self):
    """
    Rounds the pixel values in an image to integer values, while preserving the type of the data.

    At certain stages in the astronomical image generation process, detectors effectively round to
    the nearest integer.  The exact stage at which this happens depends on the type of device (CCD
    vs. NIR detector).  For example, for H2RG detectors, quantization happens in two stages: first,
    when detecting a certain number of photons, corresponding to the sum of background and signal
    multiplied by the QE and including reciprocity failure.  After this, a number of other processes
    occur (e.g., nonlinearity, IPC, read noise) that could result in non-integer pixel values, only
    rounding to an integer at the stage of analog-to-digital conversion.

    Because we cannot guarantee that quantization will always be the last step in the process, the
    quantize() routine does not actually modify the type of the image to 'int'.  However, users can
    easily do so by doing::

        >>> image.quantize()
        >>> int_image = galsim.Image(image, dtype=int)

    """
    self.applyNonlinearity(np.round)


Image.applyNonlinearity = applyNonlinearity
Image.addReciprocityFailure = addReciprocityFailure
Image.applyIPC = applyIPC
Image.applyPersistence = applyPersistence
Image.quantize = quantize



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os
import sys

import galsim
from galsim_test_helpers import *


@timer
def test_vk(run_slow):
    """Test the generation of VonKarman profiles
    """
    if run_slow:
        lams = [300.0, 500.0, 1100.0]
        r0_500s = [0.05, 0.15, 0.3]
        L0s = [1e10, 25.0, 10.0]
        do_deltas = [False, True]
    else:
        lams = [500.0]
        r0_500s = [0.2]
        L0s = [25.0]
        do_deltas = [False]
    for lam in lams:
        for r0_500 in r0_500s:
            r0 = r0_500*(lam/500)**(6./5)
            for L0 in L0s:
                for do_delta in do_deltas:
                    kwargs = {'lam':lam, 'r0':r0, 'L0':L0, 'do_delta':do_delta}
                    print(kwargs)
                    delta_amp = np.exp(-0.5*0.172629*(r0/L0)**(-5./3.))
                    if delta_amp > 1.e-3:
                        print("Skip this combination, since delta > maxk_threshold")
                        continue

                    vk = galsim.VonKarman(flux=2.2, **kwargs)
                    np.testing.assert_almost_equal(vk.flux, 2.2)

                    gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
                    vk2 = galsim.VonKarman(flux=2.2, gsparams=gsp, **kwargs)
                    assert vk2 != vk
                    assert vk2 == vk.withGSParams(gsp)
                    assert vk2 == vk.withGSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)

                    check_basic(vk, "VonKarman")
                    check_pickle(vk)

                    img = galsim.Image(16, 16, scale=0.25)
                    if not do_delta:
                        do_shoot(vk, img, "VonKarman")
                        do_kvalue(vk, img, "VonKarman")

    with np.testing.assert_raises(galsim.GalSimIncompatibleValuesError):
        vk = galsim.VonKarman(lam=500, r0=0.1, r0_500=0.2)
    with np.testing.assert_raises(galsim.GalSimIncompatibleValuesError):
        vk = galsim.VonKarman(lam=500)


@timer
def test_vk_delta():
    """Test a VonKarman with a significant delta-function amplitude"""
    kwargs = {'lam':1100.0, 'r0':0.8, 'L0':5.0, 'flux':2.2}
    # Try to see if we can catch the warning first
    with assert_warns(galsim.GalSimWarning):
        vk = galsim.VonKarman(**kwargs)

    kwargs['suppress_warning'] = True
    vk = galsim.VonKarman(**kwargs)
    check_pickle(vk)

    # This profile has more than 15% of its flux in the delta-function component.
    assert vk.delta_amplitude > 0.15 * vk.flux
    # If do_delta is False (the default), then the asymptotic kValue should still be zero.
    np.testing.assert_almost_equal(vk.kValue(1e10, 0).real, 0.0)
    # But if we use do_delta=True, then the asymptotic kValue should be that of the delta function.
    vkd = galsim.VonKarman(do_delta=True, **kwargs)
    check_pickle(vkd)
    np.testing.assert_almost_equal(vkd.kValue(1e10, 0).real, vkd.delta_amplitude)

    # Either way, the fluxes should be the same.
    np.testing.assert_almost_equal(vk.flux, vkd.flux)
    assert vk != vkd
    # The half-light-radius of the profile with do_delta=True should be smaller though, as we're
    # accounting for the 15% flux at r=0 in this case
    assert vkd.half_light_radius < vk.half_light_radius


@timer
def test_vk_scale():
    """Test vk scale argument"""
    kwargs = {'lam':500, 'r0':0.2, 'L0':25.0, 'flux':2.2}
    vk_arcsec = galsim.VonKarman(scale_unit=galsim.arcsec, **kwargs)
    vk_arcmin = galsim.VonKarman(scale_unit='arcmin', **kwargs)
    check_pickle(vk_arcmin)

    np.testing.assert_almost_equal(vk_arcsec.flux, vk_arcmin.flux)
    np.testing.assert_almost_equal(vk_arcsec.kValue(0.0, 0.0), vk_arcmin.kValue(0.0, 0.0))
    np.testing.assert_almost_equal(vk_arcsec.kValue(0.0, 10.0), vk_arcmin.kValue(0.0, 600.0))
    np.testing.assert_almost_equal(vk_arcsec.xValue(0.0, 6.0), vk_arcmin.xValue(0.0, 0.1))

    img1 = vk_arcsec.drawImage(nx=32, ny=32, scale=0.2)
    img2 = vk_arcmin.drawImage(nx=32, ny=32, scale=0.2/60.0)
    np.testing.assert_almost_equal(img1.array, img2.array)


@timer
def test_vk_properties():
    """Test some basic properties of the VonKarman profile.
    """
    # Regression test based on v2.3.5 version of the code.

    test_flux = 1.8
    psf = galsim.VonKarman(lam=500, r0=0.2, L0=25.0, flux=test_flux)

    # Check various properties
    np.testing.assert_equal(psf.centroid, galsim.PositionD(0,0))
    np.testing.assert_almost_equal(psf.maxk, 24.511275061996837)
    np.testing.assert_almost_equal(psf.stepk, 1.1025979141287368, decimal=6)
    np.testing.assert_almost_equal(psf.kValue(0,0), test_flux+0j)
    np.testing.assert_almost_equal(psf.xValue(0,0), 7.91805413536067)
    np.testing.assert_almost_equal(psf.kValue(0,0), (1+0j) * test_flux)
    np.testing.assert_almost_equal(psf.flux, test_flux)
    np.testing.assert_almost_equal(psf.xValue(0,0), psf.max_sb)

    # Check input flux vs output flux
    for inFlux in np.logspace(-2, 2, 10):
        psfFlux = galsim.VonKarman(lam=500, r0=0.2, L0=25.0, flux=inFlux)
        outFlux = psfFlux.flux
        np.testing.assert_almost_equal(outFlux, inFlux)

    # Check that stepk and maxk scale correctly with r0,L0
    psf2 = galsim.VonKarman(lam=500, r0=0.2/5, L0=25.0/5, flux=test_flux)
    np.testing.assert_almost_equal(psf2.maxk, psf.maxk/5)
    # This is not exact, since it's a computed quantity from the lookup table, so only
    # equal to about 1%.
    np.testing.assert_allclose(psf2.stepk, psf.stepk/5, rtol=0.01)

    # Equivalent if scale lam instead.
    psf2 = galsim.VonKarman(lam=5*500, r0=0.2, L0=25.0, flux=test_flux)
    np.testing.assert_almost_equal(psf2.maxk, psf.maxk/5)
    np.testing.assert_allclose(psf2.stepk, psf.stepk/5, rtol=0.01)


@timer
def test_vk_shoot():
    """Test VonKarman with photon shooting.  Particularly the flux of the final image.
    """
    rng = galsim.BaseDeviate(1234)
    obj = galsim.VonKarman(lam=500, r0=0.2, flux=1.e4)
    im = galsim.Image(100,100, scale=1)
    im.setCenter(0,0)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng)
    assert photons2 == photons, "VonKarman makePhot not equivalent to drawPhot"

    obj = galsim.VonKarman(lam=500, r0=0.2, L0=10., flux=1.e4)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng)
    assert photons2 == photons, "VonKarman makePhot not equivalent to drawPhot"

    obj = galsim.VonKarman(lam=700, r0=0.02, L0=10., flux=1.e4)
    im = galsim.Image(500,500, scale=1)
    im.setCenter(0,0)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng.duplicate())
    assert photons2 == photons, "VonKarman makePhot not equivalent to drawPhot"

    # Can treat the profile as a convolution of a delta function and put it in a photon_ops list.
    delta = galsim.DeltaFunction(flux=1.e4)
    psf = galsim.VonKarman(lam=700, r0=0.02, L0=10.)
    photons3 = delta.makePhot(poisson_flux=False, rng=rng.duplicate(), photon_ops=[psf])
    assert photons3 == photons, "Using VonKarman in photon_ops not equivalent to drawPhot"


@timer
def test_vk_ne():
    gsp = galsim.GSParams(maxk_threshold=1.1e-3, folding_threshold=5.1e-3)

    objs = [galsim.VonKarman(lam=500.0, r0=0.2),
            galsim.VonKarman(lam=500.0, r0=0.2, L0=20.0),
            galsim.VonKarman(lam=500.0, r0=0.2, L0=20.0, flux=2.2),
            galsim.VonKarman(lam=500.0, r0=0.2, L0=1e11),
            galsim.VonKarman(lam=550.0, r0=0.1, L0=20.0),
            galsim.VonKarman(lam=550.0, r0=0.1, L0=20.0, do_delta=True),
            galsim.VonKarman(lam=550.0, r0=0.1, L0=20.0, scale_unit=galsim.arcmin),
            galsim.VonKarman(lam=550.0, r0=0.1, L0=20.0, gsparams=gsp),
            galsim.VonKarman(lam=550.0, r0=0.1, L0=20.0, gsparams=gsp, force_stepk=1.0)]
    check_all_diff(objs)


@timer
def test_vk_eq_kolm():
    lam = 500.0
    r0 = 0.2
    L0 = 1e10  # Need to make this surprisingly large to make vk -> kolm.
    flux = 3.3
    kolm = galsim.Kolmogorov(lam=lam, r0=r0, flux=flux)
    vk = galsim.VonKarman(lam=lam, r0=r0, L0=L0, flux=flux)

    np.testing.assert_allclose(kolm.xValue(0,0), vk.xValue(0,0), rtol=5e-4, atol=0)

    kolm_img = kolm.drawImage(nx=24, ny=24, scale=0.2)
    vk_img = vk.drawImage(nx=24, ny=24, scale=0.2)
    np.testing.assert_allclose(kolm_img.array, vk_img.array, atol=flux*4e-5, rtol=0)


@timer
def test_vk_fitting_formulae():
    #         lam, r0_500, L0
    params = [(650, 0.15, 10.0),
              (450, 0.12, 25.0),
              (900, 0.18, 100.0)]

    def predicted_FWHM_ratio(r0, L0):
        """Fitting formula for VonKarman FWHM / Kolmogorov FWHM
        from Martinez++2014
        """
        return np.sqrt(1 - 2.183*(r0/L0)**0.356)

    def predicted_HLR_ratio(r0, L0):
        """Fitting formula for VonKarman HLR / Kolmogorov HLR
        from Martinez++2014
        """
        return np.sqrt(1 - 1.534*(r0/L0)**0.347)

    for lam, r0_500, L0 in params:
        print(lam, r0_500, L0)
        r0 = r0_500*(lam/500.0)**(6./5)
        kolm = galsim.Kolmogorov(lam=lam, r0=r0)
        vk = galsim.VonKarman(lam=lam, r0=r0, L0=L0)
        vk2 = galsim.VonKarman(lam=lam, r0_500=r0_500, L0=L0)
        np.testing.assert_allclose(vk.r0, vk2.r0)
        np.testing.assert_allclose(vk.r0_500, vk2.r0_500)
        for prof in [vk, vk2]:
            HLR_ratio = prof.calculateHLR() / kolm.calculateHLR()
            FWHM_ratio = prof.calculateFWHM() / kolm.calculateFWHM()
            print(HLR_ratio)
            print(FWHM_ratio)
            np.testing.assert_allclose(HLR_ratio, predicted_HLR_ratio(r0, L0), rtol=0.015)
            np.testing.assert_allclose(FWHM_ratio, predicted_FWHM_ratio(r0, L0), rtol=0.015)


@timer
def test_vk_gsp():
    """Test that we can construct a vK with non-standard folding_threshold.
    """
    # default folding_threshold is 5e-3.
    # We can't go too much smaller than this for such a flat asymptotic profile, but check a little
    # bit further works.
    gsp1 = galsim.GSParams(folding_threshold=1e-2)
    gsp2 = galsim.GSParams(folding_threshold=2e-3)

    # Just testing that these construct successfully
    galsim.VonKarman(lam=700, r0=0.1, L0=24.3, gsparams=gsp1)
    galsim.VonKarman(lam=700, r0=0.1, L0=24.3, gsparams=gsp2)


def vk_benchmark():
    import time
    t0 = time.time()
    vk = galsim.VonKarman(lam=700, r0=0.1, L0=24.3)
    vk.drawImage(nx=16, ny=16, scale=0.2)
    t1 = time.time()
    print("Time to create/draw first time: {:6.3f}s".format(t1-t0))  # ~0.7s
    for i in range(10):
        vk.drawImage(nx=16, ny=16, scale=0.2)
    t2 = time.time()
    print("Time to draw 10 more: {:6.3f}s".format(t2-t1))  # ~0.07s
    for i in range(100):
        vk.drawImage(nx=16, ny=16, scale=0.2, method='phot', n_photons=50000)
    t3 = time.time()
    print("Time to photon-shoot 100 more with 50000 photons each: {:6.3f}s".format(t3-t2))  # ~0.9s


@timer
def test_vk_r0(run_slow):
    """Test a special r0 value that resulted in an error, reported in issue #957.

    Note: the resolution of the bug was to add explicit split points for the first several
    j0 zeros.  Without that, the integral in rawXValue can spuriously fail badly, leading to
    an invalid estimate of the total integrated flux within R=pi/stepk.

    Update: With the new Ogata method for doing the Hankel transform, this seems no longer to
    be necessary.  However, we continue to test these r values anyway.
    """
    # The first one was issue #957.
    # Aaron Roodman ran across another, which is now included here as well.
    r0_list = [0.146068884, 0.16879518207956518]

    for r0 in r0_list:
        vk = galsim.VonKarman(L0=25.,lam=700.,r0=r0)
        check_basic(vk, "VonKarman, r0=%s"%r0)

    if run_slow:
        # Josh then tried a bunch more random triples of r0_500, lam, L0 to find more failures,
        # which are given in input/vk_fail.txt.
        r0_500_list, lam_list, L0_list = np.loadtxt('input/vk_fail.txt').T
        for r0_500, lam, L0 in zip(r0_500_list, lam_list, L0_list):
            print(r0_500,lam,L0)
            vk = galsim.VonKarman(L0=L0,lam=lam,r0_500=r0_500)
            #check_basic(vk, "VonKarman, r0_500=%s"%r0_500)


@timer
def test_vk_force_stepk():
    """Check that manually forcing stepk works"""
    vk1 = galsim.VonKarman(r0_500=0.1, L0=25.0, lam=750.0)
    vk2 = galsim.VonKarman(r0_500=0.1, L0=25.0, lam=750.0, force_stepk=10.0)

    # Make sure we get expected stepk
    assert vk1.stepk != vk2.stepk
    assert vk2.stepk == 10.0

    # Many products will actually be the same for both
    # Asking for the half_light_radius or xValue will trigger the table build,
    # which is identical for each.
    assert vk1.half_light_radius == vk2.half_light_radius
    assert vk1.xValue(0, 1) == vk2.xValue(0, 1)

    # Images will be the same if you assert specific bounds
    img1 = vk1.drawImage(nx=50, ny=50, scale=0.2, method='fft')
    img2 = vk2.drawImage(nx=50, ny=50, scale=0.2, method='fft')

    np.testing.assert_equal(img1.array, img2.array)

    # Though "goodImageSize" will differ.
    assert vk1.getGoodImageSize(0.2) != vk2.getGoodImageSize(0.2)

    # Can we pickle?
    check_pickle(vk2)
    check_pickle(vk2, lambda obj:obj.stepk)
    check_basic(vk2, 'vk2', do_x=False)  # x fails b/c stamp size is bad

    img = galsim.Image(50, 50, scale=0.2)
    do_shoot(vk2, img, "VonKarman")

    # Check works with scale
    vk3 = galsim.VonKarman(
        r0_500=0.1, L0=25.0, lam=750.0, force_stepk=10.0,
        scale_unit=galsim.radians
    )
    assert vk3.stepk == 10.0
    assert vk3.scale_unit == galsim.radians

    # force_stepk is retained through a reflux
    vk4 = vk3.withFlux(11.0)
    assert vk4.flux == 11.0
    assert vk3.force_stepk == vk4.force_stepk

@timer
def test_low_folding_threshold():
    """Test VonKarman with a very low folding_threshold.
    """
    folding_threshold = 1e-4
    pixel_scale = 0.2
    kwargs = {'lam':500, 'r0':0.2, 'L0':25.0, 'flux':2.2}
    gsparams = galsim.GSParams(folding_threshold=folding_threshold)
    psf = galsim.VonKarman(gsparams=gsparams, **kwargs)
    image_size = psf.getGoodImageSize(pixel_scale)
    print('ft = 1.e-4: psf.getGoodImageSize:', image_size)
    assert image_size == 298

    folding_threshold = 1e-6
    gsparams = galsim.GSParams(folding_threshold=folding_threshold)
    psf = galsim.VonKarman(gsparams=gsparams, **kwargs)
    image_size = psf.getGoodImageSize(pixel_scale)
    print('ft = 1.e-6: psf.getGoodImageSize:', image_size)
    assert image_size == 600

    # Check an extremely small L0
    kwargs['L0'] = 1.0
    with assert_warns(galsim.GalSimWarning):
        # This low an L0 has a non-negligible delta function, hence a warning.
        psf = galsim.VonKarman(gsparams=gsparams, **kwargs)
    image_size = psf.getGoodImageSize(pixel_scale)
    print('L0=1.0, ft = 1.e-6: psf.getGoodImageSize:', image_size)
    assert image_size == 600


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("--benchmark", action='store_true', help="Run timing benchmark")

    runtests(__file__, parser=parser)

    args, unknown_args = parser.parse_known_args()
    if args.benchmark:
        vk_benchmark()



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Gaussian' ]

import numpy as np
import math

from . import _galsim
from .gsobject import GSObject
from .gsparams import GSParams
from .utilities import lazy_property, doc_inherit
from .errors import GalSimIncompatibleValuesError


class Gaussian(GSObject):
    r"""A class describing a 2D Gaussian surface brightness profile.

    The Gaussian surface brightness profile is characterized by two properties, its ``flux``
    and the characteristic size ``sigma`` where the radial profile of the circular Gaussian
    drops off as

    .. math::
        I(r) \sim e^{-\frac{r^2}{2 \sigma^2}}

    A Gaussian can be initialized using one (and only one) of three possible size parameters:
    ``sigma``, ``fwhm``, or ``half_light_radius``.  Exactly one of these three is required.

    Parameters:
        sigma:              The value of sigma of the profile.  Typically given in arcsec.
                            [One of ``sigma``, ``fwhm``, or ``half_light_radius`` is required.]
        fwhm:               The full-width-half-max of the profile.  Typically given in arcsec.
                            [One of ``sigma``, ``fwhm``, or ``half_light_radius`` is required.]
        half_light_radius:  The half-light radius of the profile.  Typically given in arcsec.
                            [One of ``sigma``, ``fwhm``, or ``half_light_radius`` is required.]
        flux:               The flux (in photons/cm^2/s) of the profile. [default: 1]
        gsparams:           An optional `GSParams` argument. [default: None]
    """
    _opt_params = { "flux" : float }
    _single_params = [ { "sigma" : float, "half_light_radius" : float, "fwhm" : float } ]

    # The FWHM of a Gaussian is 2 sqrt(2 ln2) sigma
    _fwhm_factor = 2.3548200450309493
    # The half-light-radius is sqrt(2 ln2) sigma
    _hlr_factor = 1.1774100225154747
    # 1/(2pi)
    _inv_twopi = 0.15915494309189535

    _has_hard_edges = False
    _is_axisymmetric = True
    _is_analytic_x = True
    _is_analytic_k = True

    def __init__(self, half_light_radius=None, sigma=None, fwhm=None, flux=1., gsparams=None):
        if fwhm is not None :
            if sigma is not None or half_light_radius is not None:
                raise GalSimIncompatibleValuesError(
                    "Only one of sigma, fwhm, and half_light_radius may be specified",
                    fwhm=fwhm, sigma=sigma, half_light_radius=half_light_radius)
            else:
                sigma = fwhm / Gaussian._fwhm_factor
        elif half_light_radius is not None:
            if sigma is not None:
                raise GalSimIncompatibleValuesError(
                    "Only one of sigma, fwhm, and half_light_radius may be specified",
                    fwhm=fwhm, sigma=sigma, half_light_radius=half_light_radius)
            else:
                sigma = half_light_radius / Gaussian._hlr_factor
        elif sigma is None:
            raise GalSimIncompatibleValuesError(
                "One of sigma, fwhm, and half_light_radius must be specified",
                fwhm=fwhm, sigma=sigma, half_light_radius=half_light_radius)

        self._sigma = float(sigma)
        self._flux = float(flux)
        self._gsparams = GSParams.check(gsparams)
        self._sigsq = sigma**2
        self._inv_sigsq = 1./self._sigsq
        self._norm = self.flux * self._inv_sigsq * Gaussian._inv_twopi

    @lazy_property
    def _sbp(self):
        return _galsim.SBGaussian(self._sigma, self._flux, self.gsparams._gsp)

    @property
    def sigma(self):
        """The sigma of this Gaussian profile
        """
        return self._sigma

    @property
    def half_light_radius(self):
        """The half-light radius of this Gaussian profile
        """
        return self.sigma * Gaussian._hlr_factor

    @property
    def fwhm(self):
        """The FWHM of this Gaussian profile
        """
        return self.sigma * Gaussian._fwhm_factor

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, Gaussian) and
                 self.sigma == other.sigma and
                 self.flux == other.flux and
                 self.gsparams == other.gsparams))

    def __hash__(self):
        return hash(("galsim.Gaussian", self.sigma, self.flux, self.gsparams))

    def __repr__(self):
        return 'galsim.Gaussian(sigma=%r, flux=%r, gsparams=%r)'%(
            self.sigma, self.flux, self.gsparams)

    def __str__(self):
        s = 'galsim.Gaussian(sigma=%s'%self.sigma
        if self.flux != 1.0:
            s += ', flux=%s'%self.flux
        s += ')'
        return s

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def _maxk(self):
        return math.sqrt(-2.*math.log(self.gsparams.maxk_threshold))/self.sigma

    @property
    def _stepk(self):
        R = max(math.sqrt(-2.*math.log(self.gsparams.folding_threshold)),
                self.gsparams.stepk_minimum_hlr * Gaussian._hlr_factor)
        return math.pi / (R * self.sigma)

    @property
    def _max_sb(self):
        return self._norm

    def _xValue(self, pos):
        rsq = pos.x**2 + pos.y**2
        return self._norm * math.exp(-0.5 * rsq * self._inv_sigsq)

    def _kValue(self, kpos):
        ksq = (kpos.x**2 + kpos.y**2) * self._sigsq
        return self._flux * math.exp(-0.5 * ksq)

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        dx,dy = offset
        self._sbp.draw(image._image, image.scale, _jac, dx, dy, flux_scaling)

    def _shoot(self, photons, rng):
        self._sbp.shoot(photons._pa, rng._rng)

    def _drawKImage(self, image, jac=None):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        self._sbp.drawK(image._image, image.scale, _jac)

    @doc_inherit
    def withFlux(self, flux):
        return Gaussian(sigma=self.sigma, flux=flux, gsparams=self.gsparams)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os

import galsim
from galsim_test_helpers import *

path, filename = os.path.split(__file__)
imgdir = os.path.join(path, "SBProfile_comparison_images") # Directory containing the reference
                                                           # images.

@timer
def test_shapelet_gaussian():
    """Test that the simplest Shapelet profile is equivalent to a Gaussian
    """
    ftypes = [np.float32, np.float64]
    scale = 0.2
    test_flux = 23.

    # First, a Shapelet with only b_00 = 1 should be identically a Gaussian
    im1 = galsim.ImageF(64,64, scale=scale)
    im2 = galsim.ImageF(64,64, scale=scale)
    for sigma in [1., 0.6, 2.4]:
        gauss = galsim.Gaussian(flux=test_flux, sigma=sigma)
        gauss.drawImage(im1, method='no_pixel')
        for order in [0, 2, 8]:
            bvec = np.zeros(galsim.Shapelet.size(order))
            bvec[0] = test_flux
            shapelet = galsim.Shapelet(sigma=sigma, order=order, bvec=bvec)
            shapelet.drawImage(im2, method='no_pixel')
            printval(im2,im1)
            np.testing.assert_array_almost_equal(
                    im1.array, im2.array, 5,
                    err_msg="Shapelet with (only) b00=1 disagrees with Gaussian result"
                            "for flux=%f, sigma=%f, order=%d"%(test_flux,sigma,order))
            np.testing.assert_almost_equal(
                    gauss.max_sb, shapelet.max_sb, 5,
                    err_msg="Shapelet max_sb did not match Gaussian max_sb")
            np.testing.assert_almost_equal(
                    gauss.flux, shapelet.flux, 5,
                    err_msg="Shapelet flux did not match Gaussian flux")


@timer
def test_shapelet_drawImage():
    """Test some measured properties of a drawn shapelet against the supposed true values
    """
    ftypes = [np.float32, np.float64]
    scale = 0.2
    test_flux = 23.

    im = galsim.ImageF(129,129, scale=scale)
    for sigma in [1., 0.3, 2.4]:
        for order in [0, 2, 8]:
            bvec = np.zeros(galsim.Shapelet.size(order))
            bvec[0] = 1.  # N,m = 0,0
            k = 0
            for n in range(1,order+1):
                k += n+1
                if n%2 == 0:  # even n
                    bvec[k] = 0.23/(n*n)        # N,m = n,0  or p,q = n/2,n/2
                    if n >= 2:
                        bvec[k-2] = 0.14/n      # N,m = n,2  real part
                        bvec[k-1] = -0.08/n     # N,m = n,2  imag part
                else:  # odd n
                    if n >= 1:
                        bvec[k-1] = -0.08/n**3.2    # N,m = n,1  real part
                        bvec[k] = 0.05/n**2.1       # N,m = n,1  imag part
                    if n >= 3:
                        bvec[k-3] = 0.31/n**4.2    # N,m = n,3  real part
                        bvec[k-2] = -0.18/n**3.9       # N,m = n,3  imag part
            print('shapelet vector = ',bvec)
            shapelet = galsim.Shapelet(sigma=sigma, order=order, bvec=bvec)

            gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
            shapelet2 = galsim.Shapelet(sigma=sigma, order=order, bvec=bvec, gsparams=gsp)
            assert shapelet2 != shapelet
            assert shapelet2 == shapelet.withGSParams(gsp)
            assert shapelet2 == shapelet.withGSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)

            check_basic(shapelet, "Shapelet", approx_maxsb=True)

            # Test normalization  (This is normally part of do_shoot.  When we eventually
            # implement photon shooting, we should go back to the normal do_shoot call,
            # and remove this section.)
            shapelet = shapelet.withFlux(test_flux)
            shapelet.drawImage(im)
            flux = im.array.sum()
            print('im.sum = ',flux,'  cf. ',test_flux)
            np.testing.assert_almost_equal(flux / test_flux, 1., 4,
                    err_msg="Flux normalization for Shapelet disagrees with expected result")
            np.testing.assert_allclose(
                    im.array.max(), shapelet.max_sb * im.scale**2, rtol=0.1,
                    err_msg="Shapelet max_sb did not match maximum pixel")

            # Test centroid
            # Note: this only works if the image has odd sizes.  If they are even, then
            # setCenter doesn't actually set the center to the true center of the image
            # (since it falls between pixels).
            im.setCenter(0,0)
            x,y = np.meshgrid(np.arange(im.array.shape[0]).astype(float) + im.xmin,
                              np.arange(im.array.shape[1]).astype(float) + im.ymin)
            x *= scale
            y *= scale
            flux = im.array.sum()
            mx = (x*im.array).sum() / flux
            my = (y*im.array).sum() / flux
            conv = galsim.Convolve([shapelet, galsim.Pixel(scale)])
            print('centroid = ',mx,my,' cf. ',conv.centroid)
            np.testing.assert_almost_equal(mx, shapelet.centroid.x, 3,
                    err_msg="Measured centroid (x) for Shapelet disagrees with expected result")
            np.testing.assert_almost_equal(my, shapelet.centroid.y, 3,
                    err_msg="Measured centroid (y) for Shapelet disagrees with expected result")


@timer
def test_shapelet_properties():
    """Test some specific numbers for a particular Shapelet profile.
    """
    # A semi-random particular vector of coefficients.
    sigma = 1.8
    order = 4
    bvec = [1.3,                               # n = 0
            0.02, 0.03,                        # n = 1
            0.23, -0.19, 0.08,                 # n = 2
            0.01, 0.02, 0.04, -0.03,           # n = 3
            -0.09, 0.07, -0.11, -0.08, 0.11]   # n = 4

    shapelet = galsim.Shapelet(sigma=sigma, order=order, bvec=bvec)

    assert shapelet.sigma == sigma
    assert shapelet.order == order
    np.testing.assert_array_equal(shapelet.bvec, bvec)

    check_basic(shapelet, "Shapelet", approx_maxsb=True)

    # Check flux
    flux = bvec[0] + bvec[5] + bvec[14]
    np.testing.assert_almost_equal(shapelet.flux, flux, 10)
    # The max_sb is not very accurate for Shapelet, but in this case it is still ok (matching
    # xValue(0,0), which isn't actually the maximum) to 2 digits.
    np.testing.assert_almost_equal(
            shapelet.xValue(0,0), shapelet.max_sb, 2,
            err_msg="Shapelet max_sb did not match maximum pixel value")
    # Check centroid
    cen = galsim.PositionD(bvec[1],-bvec[2]) + np.sqrt(2.) * galsim.PositionD(bvec[8],-bvec[9])
    cen *= 2. * sigma / flux
    np.testing.assert_almost_equal(shapelet.centroid.x, cen.x, 10)
    np.testing.assert_almost_equal(shapelet.centroid.y, cen.y, 10)
    # Check Fourier properties
    np.testing.assert_almost_equal(shapelet.maxk, 4.61738371186, 10)
    np.testing.assert_almost_equal(shapelet.stepk, 0.195133742529, 10)
    # Check image values in real and Fourier space
    zero = galsim.PositionD(0., 0.)
    np.testing.assert_almost_equal(shapelet.kValue(zero), flux+0j, 10)
    np.testing.assert_almost_equal(shapelet.xValue(zero), 0.0653321217013, 10)

    # Check picklability
    check_pickle(shapelet)

    assert_raises(TypeError, galsim.Shapelet, sigma=sigma)
    assert_raises(TypeError, galsim.Shapelet, sigma=sigma, bvec=bvec)
    assert_raises(TypeError, galsim.Shapelet, order=order, bvec=bvec)
    assert_raises(ValueError, galsim.Shapelet, sigma=sigma, order=5, bvec=bvec)

    # Check that stepk and maxk scale correctly with radius
    shapelet2 = galsim.Shapelet(sigma=5*sigma, order=order, bvec=bvec)
    np.testing.assert_almost_equal(shapelet2.maxk, shapelet.maxk/5)
    np.testing.assert_almost_equal(shapelet2.stepk, shapelet.stepk/5)

@timer
def test_shapelet_fit():
    """Test fitting a Shapelet decomposition of an image
    """
    for method, norm in [('no_pixel','f'), ('sb','sb')]:
        # We fit a shapelet approximation of a distorted Moffat profile:
        flux = 20
        psf = galsim.Moffat(beta=3.4, half_light_radius=1.2, flux=flux)
        psf = psf.shear(g1=0.11,g2=0.07).shift(0.03,0.04)
        scale = 0.2
        pixel = galsim.Pixel(scale)
        conv = galsim.Convolve([psf,pixel])
        im1 = conv.drawImage(scale=scale, method=method)

        sigma = 1.2  # Match half-light-radius as a decent first approximation.
        shapelet = galsim.Shapelet.fit(sigma, 10, im1, normalization=norm)
        print('fitted shapelet coefficients = ',shapelet.bvec)

        # Check flux
        print('flux = ',shapelet.flux,'  cf. ',flux)
        np.testing.assert_almost_equal(shapelet.flux / flux, 1., 1,
                err_msg="Fitted shapelet has the wrong flux")

        # Test centroid
        print('centroid = ',shapelet.centroid,'  cf. ',conv.centroid)
        np.testing.assert_almost_equal(shapelet.centroid.x, conv.centroid.x, 2,
                err_msg="Fitted shapelet has the wrong centroid (x)")
        np.testing.assert_almost_equal(shapelet.centroid.y, conv.centroid.y, 2,
                err_msg="Fitted shapelet has the wrong centroid (y)")

        # Test drawing image from shapelet
        im2 = im1.copy()
        shapelet.drawImage(im2, method=method)
        # Check that images are close to the same:
        print('norm(diff) = ',np.sum((im1.array-im2.array)**2))
        print('norm(im) = ',np.sum(im1.array**2))
        print('max(diff) = ',np.max(np.abs(im1.array-im2.array)))
        print('max(im) = ',np.max(np.abs(im1.array)))
        peak_scale = np.max(np.abs(im1.array))*3  # Check agreement to within 3% of peak value.
        np.testing.assert_almost_equal(im2.array/peak_scale, im1.array/peak_scale, decimal=2,
                err_msg="Shapelet version not a good match to original")

        # Remeasure -- should now be very close to the same.
        shapelet2 = galsim.Shapelet.fit(sigma, 10, im2, normalization=norm)
        np.testing.assert_equal(shapelet.sigma, shapelet2.sigma,
                err_msg="Second fitted shapelet has the wrong sigma")
        np.testing.assert_equal(shapelet.order, shapelet2.order,
                err_msg="Second fitted shapelet has the wrong order")
        np.testing.assert_almost_equal(shapelet.bvec, shapelet2.bvec, 6,
                err_msg="Second fitted shapelet coefficients do not match original")

        # Test drawing off center
        im2 = im1.copy()
        offset = galsim.PositionD(0.3,1.4)
        shapelet.drawImage(im2, method=method, offset=offset)
        shapelet2 = galsim.Shapelet.fit(sigma, 10, im2, normalization=norm,
                                        center=im2.true_center + offset)
        np.testing.assert_equal(shapelet.sigma, shapelet2.sigma,
                err_msg="Second fitted shapelet has the wrong sigma")
        np.testing.assert_equal(shapelet.order, shapelet2.order,
                err_msg="Second fitted shapelet has the wrong order")
        np.testing.assert_almost_equal(shapelet.bvec, shapelet2.bvec, 6,
                err_msg="Second fitted shapelet coefficients do not match original")

    assert_raises(ValueError, galsim.Shapelet.fit, sigma, 10, im1, normalization='invalid')

    # Haven't gotten around to implementing this yet...
    im2.wcs = galsim.JacobianWCS(0.2,0.01,0.01,0.2)
    with assert_raises(NotImplementedError):
        galsim.Shapelet.fit(sigma, 10, im2)


@timer
def test_shapelet_adjustments():
    """Test that adjusting the Shapelet profile in various ways does the right thing
    """
    ftypes = [np.float32, np.float64]

    nx = 128
    ny = 128
    scale = 0.2
    im = galsim.ImageF(nx,ny, scale=scale)

    sigma = 1.8
    order = 6
    bvec = [1.3,                                            # n = 0
            0.02, 0.03,                                     # n = 1
            0.23, -0.19, 0.08,                              # n = 2
            0.01, 0.02, 0.04, -0.03,                        # n = 3
            -0.09, 0.07, -0.11, -0.08, 0.11,                # n = 4
            -0.03, -0.02, -0.08, 0.01, -0.06, -0.03,        # n = 5
            0.06, -0.02, 0.00, -0.05, -0.04, 0.01, 0.09 ]   # n = 6

    ref_shapelet = galsim.Shapelet(sigma=sigma, order=order, bvec=bvec)
    ref_im = galsim.ImageF(nx,ny)
    ref_shapelet.drawImage(ref_im, scale=scale)

    # Test PQ and NM access
    np.testing.assert_equal(ref_shapelet.getPQ(0,0), (bvec[0],0))
    np.testing.assert_equal(ref_shapelet.getPQ(1,1), (bvec[5],0))
    np.testing.assert_equal(ref_shapelet.getPQ(1,2), (bvec[8],-bvec[9]))
    np.testing.assert_equal(ref_shapelet.getPQ(3,2), (bvec[19],bvec[20]))
    np.testing.assert_equal(ref_shapelet.getNM(0,0), (bvec[0],0))
    np.testing.assert_equal(ref_shapelet.getNM(2,0), (bvec[5],0))
    np.testing.assert_equal(ref_shapelet.getNM(3,-1), (bvec[8],-bvec[9]))
    np.testing.assert_equal(ref_shapelet.getNM(5,1), (bvec[19],bvec[20]))

    # Test that the Shapelet withFlux does the same thing as the GSObject withFlux
    # Make it opaque to the Shapelet versions
    alt_shapelet = ref_shapelet + 0. * galsim.Gaussian(sigma=1)
    alt_shapelet.withFlux(23.).drawImage(ref_im, method='no_pixel')
    shapelet = galsim.Shapelet(sigma=sigma, order=order, bvec=bvec)
    shapelet.withFlux(23.).drawImage(im, method='no_pixel')
    np.testing.assert_array_almost_equal(
        im.array, ref_im.array, 6,
        err_msg="Shapelet withFlux disagrees with GSObject withFlux")

    # Test that scaling the Shapelet flux does the same thing as the GSObject scaling
    (alt_shapelet * 0.23).drawImage(ref_im, method='no_pixel')
    (shapelet * 0.23).drawImage(im, method='no_pixel')
    np.testing.assert_array_almost_equal(
        im.array, ref_im.array, 6,
        err_msg="Shapelet *= 0.23 disagrees with GSObject *= 0.23")

    # Test that the Shapelet rotate does the same thing as the GSObject rotate
    alt_shapelet.rotate(23. * galsim.degrees).drawImage(ref_im, method='no_pixel')
    shapelet.rotate(23. * galsim.degrees).drawImage(im, method='no_pixel')
    np.testing.assert_array_almost_equal(
        im.array, ref_im.array, 6,
        err_msg="Shapelet rotate disagrees with GSObject rotate")

    # Test that the Shapelet dilate does the same thing as the GSObject dilate
    alt_shapelet.dilate(1.3).drawImage(ref_im, method='no_pixel')
    shapelet.dilate(1.3).drawImage(im, method='no_pixel')
    np.testing.assert_array_almost_equal(
        im.array, ref_im.array, 6,
        err_msg="Shapelet dilate disagrees with GSObject dilate")

    # Test that the Shapelet expand does the same thing as the GSObject expand
    alt_shapelet.expand(1.7).drawImage(ref_im, method='no_pixel')
    shapelet.expand(1.7).drawImage(im, method='no_pixel')
    np.testing.assert_array_almost_equal(
        im.array, ref_im.array, 6,
        err_msg="Shapelet expand disagrees with GSObject expand")

    # Test that the Shapelet magnify does the same thing as the GSObject magnify
    alt_shapelet.magnify(0.8).drawImage(ref_im, method='no_pixel')
    shapelet.magnify(0.8).drawImage(im, method='no_pixel')
    np.testing.assert_array_almost_equal(
        im.array, ref_im.array, 6,
        err_msg="Shapelet magnify disagrees with GSObject magnify")

    # Test that lens works on Shapelet
    alt_shapelet.lens(-0.05, 0.15, 1.1).drawImage(ref_im, method='no_pixel')
    shapelet.lens(-0.05, 0.15, 1.1).drawImage(im, method='no_pixel')
    np.testing.assert_array_almost_equal(
        im.array, ref_im.array, 6,
        err_msg="Shapelet lens disagrees with GSObject lens")


@timer
def test_ne():
    """ Check that inequality works as expected."""
    gsp = galsim.GSParams(maxk_threshold=5.1e-3, folding_threshold=1.1e-3)
    objs = [galsim.Shapelet(1., 2),
            galsim.Shapelet(1., 3),
            galsim.Shapelet(2., 2),
            galsim.Shapelet(1., 2, bvec=[1, 0, 0, 0.2, 0.3, -0.1]),
            galsim.Shapelet(1., 2, gsparams=gsp)]
    check_all_diff(objs)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os
from multiprocessing.managers import ListProxy, DictProxy

from .util import LoggerWrapper, SetDefaultExt, RetryIO, SafeManager, single_threaded
from .value import ParseValue
from .image import GetNObjForImage
from ..utilities import ensure_dir
from ..errors import GalSimConfigValueError, GalSimConfigError
from ..fits import writeMulti

# This file handles the processing of extra output items in addition to the primary output file
# in config['output']. The ones that are defined natively in GalSim are psf, weight, badpix,
# and truth.  See extra_*.py for the specific functions for each of these.

# This module-level dict will store all the registered "extra" output types.
# See the RegisterExtraOutput function at the end of this file.
# The keys will be the (string) names of the extra output types, and the values will be
# builder classes that will perform the different processing functions.
valid_extra_outputs = {}

def SetupExtraOutput(config, logger=None):
    """
    Set up the extra output items as necessary, including building Managers for the work
    space so they can work safely in multi-processing mode.  Each builder will be placed in
    config['extra_builder'][key] where key is the key in galsim.config.valid_extra_outputs.

    Parameters:
        config:     The configuration dict.
        logger:     If given, a logger object to log progress. [default: None]
    """
    logger = LoggerWrapper(logger)
    output = config['output']
    file_num = config.get('file_num',0)

    # We'll iterate through this list of keys a few times
    all_keys = [ k for k in valid_extra_outputs.keys() if k in output ]

    # We don't need the manager stuff if we (a) are already in a multiprocessing Process, or
    # (b) config.image.nproc == 1.
    use_manager = (
            'current_nproc' not in config and
            'image' in config and 'nproc' in config['image'] and
            ParseValue(config['image'], 'nproc', config, int)[0] != 1 )

    if use_manager and 'output_manager' not in config:
        class OutputManager(SafeManager): pass

        # We'll use a list and a dict as work space to do the extra output processing.
        OutputManager.register('dict', dict, DictProxy)
        OutputManager.register('list', list, ListProxy)
        # Start up the output_manager
        config['output_manager'] = OutputManager()
        with single_threaded():
            config['output_manager'].start()

    if 'extra_builder' not in config:
        config['extra_builder'] = {}

    # Keep track of any skipped obj_nums, since usually need to treat them differently.
    # Note: it would be slightly nicer to use a set here, but there isn't a pre-defined
    # multiprocessing.managers.SetProxy type, so we just use a dict like a set by giving
    # each item the value None.
    if '_skipped_obj_nums' in config:
        config['_skipped_obj_nums'].clear()
    elif use_manager:
        config['_skipped_obj_nums'] = config['output_manager'].dict()
    else:
        config['_skipped_obj_nums'] = dict()

    for key in all_keys:
        logger.debug('file %d: Setup output item %s',file_num,key)

        # Make the work space structures
        if use_manager:
            data = config['output_manager'].list()
            scratch = config['output_manager'].dict()
        else:
            data = list()
            scratch = dict()

        # Make the data list the right length now to avoid issues with multiple
        # processes trying to append at the same time.
        nimages = config.get('nimages', 1)
        for k in range(nimages):
            data.append(None)

        # Create the builder, giving it the data and scratch objects as work space.
        field = config['output'][key]
        builder = valid_extra_outputs[key]
        builder.initialize(data, scratch, field, config, logger)
        # And store it in the config dict
        config['extra_builder'][key] = builder

        logger.debug('file %d: Setup output %s object',file_num,key)


def SetupExtraOutputsForImage(config, logger=None):
    """Perform any necessary setup for the extra output items at the start of a new image.

    Parameters:
        config:     The configuration dict.
        logger:     If given, a logger object to log progress. [default: None]
    """
    if 'output' in config:
        if 'extra_builder' not in config:
            SetupExtraOutput(config, logger)
        for key, builder in config['extra_builder'].items():
            field = config['output'][key]
            builder.setupImage(field, config, logger)

def ProcessExtraOutputsForStamp(config, skip, logger=None):
    """Run the appropriate processing code for any extra output items that need to do something
    at the end of building each object.

    This gets called after all the object flux is added to the stamp, but before the sky level
    and noise are added.

    Parameters:
        config:     The configuration dict.
        skip:       Was the drawing of this object skipped?
        logger:     If given, a logger object to log progress. [default: None]
    """
    if 'output' in config:
        obj_num = config['obj_num']
        for key, builder in config.get('extra_builder',{}).items():
            field = config['output'][key]
            if skip:
                config['_skipped_obj_nums'][obj_num] = None
                builder.processSkippedStamp(obj_num, field, config, logger)
            else:
                builder.processStamp(obj_num, field, config, logger)


def ProcessExtraOutputsForImage(config, logger=None):
    """Run the appropriate processing code for any extra output items that need to do something
    at the end of building each image

    Parameters:
        config:     The configuration dict.
        logger:     If given, a logger object to log progress. [default: None]
    """
    if 'output' in config:
        obj_nums = None
        for key, builder in config.get('extra_builder',{}).items():
            image_num = config.get('image_num',0)
            start_image_num = config.get('start_image_num',0)
            if obj_nums is None:
                # Figure out which obj_nums were used for this image.
                file_num = config.get('file_num',0)
                start_obj_num = config.get('start_obj_num',0)
                nobj = config.get('nobj', [GetNObjForImage(config,image_num)])
                k = image_num - start_image_num
                for i in range(k):
                    start_obj_num += nobj[i]
                obj_nums = range(start_obj_num, start_obj_num+nobj[k])
                # Omit skipped obj_nums
                skipped = config['_skipped_obj_nums']
                obj_nums = [ n for n in obj_nums if n not in skipped ]
            field = config['output'][key]
            index = image_num - start_image_num
            builder.processImage(index, obj_nums, field, config, logger)


def WriteExtraOutputs(config, main_data, logger=None):
    """Write the extra output objects to files.

    This gets run at the end of the functions for building the regular output files.

    Parameters:
        config:     The configuration dict.
        main_data:  The main file data in case it is needed.
        logger:     If given, a logger object to log progress. [default: None]
    """
    logger = LoggerWrapper(logger)
    output = config['output']
    if 'retry_io' in output:
        ntries = ParseValue(config['output'],'retry_io',config,int)[0]
        # This is how many retries.  Do at least 1, so ntries is 1 more than this.
        ntries = ntries + 1
    else:
        ntries = 1

    if 'dir' in output:
        default_dir = ParseValue(output,'dir',config,str)[0]
    else:
        default_dir = None

    if 'noclobber' in output:
        noclobber = ParseValue(output,'noclobber',config,bool)[0]
    else:
        noclobber = False

    if 'extra_last_file' not in config:
        config['extra_last_file'] = {}

    for key, builder in config['extra_builder'].items():
        field = output[key]
        if 'file_name' in field:
            SetDefaultExt(field, '.fits')
            file_name = ParseValue(field,'file_name',config,str)[0]
        else:  # pragma: no cover  This is covered, but codecov thinks it isn't.
            # If no file_name, then probably writing to hdu
            continue
        if 'dir' in field:
            dir = ParseValue(field,'dir',config,str)[0]
        else:
            dir = default_dir

        if dir is not None:
            file_name = os.path.join(dir,file_name)

        ensure_dir(file_name)

        if noclobber and os.path.isfile(file_name):
            logger.warning('Not writing %s file %d = %s because output.noclobber = True '
                           'and file exists',key,config['file_num'],file_name)
            continue

        if config['extra_last_file'].get(key, None) == file_name:
            # If we already wrote this file, skip it this time around.
            # (Typically this is applicable for psf, where we may only want 1 psf file.)
            logger.info('Not writing %s file %d = %s because already written',
                        key,config['file_num'],file_name)
            continue

        # Do any final processing that needs to happen.
        builder.ensureFinalized(field, config, main_data, logger)

        # Call the write function, possibly multiple times to account for IO failures.
        write_func = builder.writeFile
        args = (file_name,field,config,logger)
        RetryIO(write_func, args, ntries, file_name, logger)
        config['extra_last_file'][key] = file_name
        logger.debug('file %d: Wrote %s to %r',config['file_num'],key,file_name)


def AddExtraOutputHDUs(config, main_data, logger=None):
    """Write the extra output objects to either HDUS or images as appropriate and add them
    to the existing data.

    This gets run at the end of the functions for building the regular output files.

    Note: the extra items must have hdu numbers ranging continuously (in any order) starting
    at len(data).  Typically first = 1, since the main image is the primary HDU, numbered 0.

    Parameters:
        config:     The configuration dict.
        main_data:  The main file data as a list of images.  Usually just [image] where
                    image is the primary image to be written to the output file.
        logger:     If given, a logger object to log progress. [default: None]

    Returns:
        data with additional hdus added
    """
    output = config['output']
    hdus = {}
    for key, builder in config['extra_builder'].items():
        field = output[key]
        if 'hdu' in field:
            hdu = ParseValue(field,'hdu',config,int)[0]
        else:  # pragma: no cover  This is covered, but codecov thinks it isn't.
            # If no hdu, then probably writing to file
            continue
        if hdu <= 0 or hdu in hdus:
            raise GalSimConfigValueError("hdu is invalid or a duplicate.",hdu)

        # Do any final processing that needs to happen.
        builder.ensureFinalized(field, config, main_data, logger)

        # Build the HDU for this output object.
        hdus[hdu] = builder.writeHdu(field,config,logger)

    first = len(main_data)
    for h in range(first,len(hdus)+first):
        if h not in hdus:
            raise GalSimConfigError("Cannot skip hdus.  No output found for hdu %d"%h)
    # Turn hdus into a list (in order)
    hdulist = [ hdus[k] for k in range(first,len(hdus)+first) ]
    return main_data + hdulist

def CheckNoExtraOutputHDUs(config, output_type, logger=None):
    """Check that none of the extra output objects want to add to the HDU list.

    Raises an exception if one of them has an hdu field.

    Parameters:
        config:         The configuration dict.
        output_type:    A string to use in the error message to indicate which output type
                        had a problem.
        logger:         If given, a logger object to log progress. [default: None]
    """
    logger = LoggerWrapper(logger)
    output = config['output']
    for key in config['extra_builder'].keys():
        field = output[key]
        if 'hdu' in field:
            hdu = ParseValue(field,'hdu',config,int)[0]
            logger.error("Extra output %s requesting to write to hdu %d", key, hdu)
            raise GalSimConfigError(
                "Output type %s cannot add extra images as HDUs"%output_type)


def GetFinalExtraOutput(key, config, main_data=[], logger=None):
    """Get the finalized output object for the given extra output key

    Parameters:
        key:        The name of the output field in config['output']
        config:     The configuration dict.
        main_data:  The main file data in case it is needed.  [default: []]
        logger:     If given, a logger object to log progress. [default: None]

    Returns:
        the final data to be output.
    """
    field = config['output'][key]
    return config['extra_builder'][key].ensureFinalized(field, config, main_data, logger)

class ExtraOutputBuilder:
    """A base class for building some kind of extra output object along with the main output.

    The base class doesn't do anything, but it defines the function signatures that a derived
    class can override to perform specific processing at any of several steps in the processing.

    The builder gets initialized with a list and and dict to use as work space.
    The typical work flow is to save something in scratch[obj_num] for each object built, and then
    process them all at the end of each image into data[k].  Then finalize may do something
    additional at the end of the processing to prepare the data to be written.

    It's worth remembering that the objects could potentially be processed in a random order if
    multiprocessing is being used.  The above work flow will thus work regardless of the order
    that the stamps and/or images are processed.

    Also, because of how objects are duplicated across processes during multiprocessing, you
    should not count on attributes you set in the builder object during the stamp or image
    processing stages to be present in the later finalize or write stages.  You should write
    any information you want to persist into the scratch or data objects, which are set up
    to handle the multiprocessing communication properly.
    """
    def initialize(self, data, scratch, config, base, logger):
        """Do any initial setup for this builder at the start of a new output file.

        The base class implementation saves two work space items into self.data and self.scratch
        that can be used to safely communicate across multiple processes.

        Parameters:
            data:       An empty list of length nimages to use as work space.
            scratch:    An empty dict that can be used as work space.
            config:     The configuration field for this output object.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress. [default: None]
        """
        self.data = data
        self.scratch = scratch
        self.final_data = None

    def setupImage(self, config, base, logger):
        """Perform any necessary setup at the start of an image.

        This function will be called at the start of each image to allow for any setup that
        needs to happen at this point in the processing.

        Parameters:
            config:     The configuration field for this output object.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress. [default: None]
        """
        pass

    def processStamp(self, obj_num, config, base, logger):
        """Perform any necessary processing at the end of each stamp construction.

        This function will be called after each stamp is built, but before the noise is added,
        so the existing stamp image has the true surface brightness profile (unless photon shooting
        was used, in which case there will necessarily be noise from that process).

        Remember, these stamps may be processed out of order.  Saving data to the scratch dict
        is safe, even if multiprocessing is being used.

        Parameters:
           obj_num:    The object number
           config:     The configuration field for this output object.
           base:       The base configuration dict.
           logger:     If given, a logger object to log progress. [default: None]
        """
        pass  # pragma: no cover  (all our ExtraBuilders override this function.)

    def processSkippedStamp(self, obj_num, config, base, logger):
        """Perform any necessary processing for stamps that were skipped in the normal processing.

        This function will be called for stamps that are not built because they were skipped
        for some reason.  Normally, you would not want to do anything for the extra outputs in
        these cases, but in case some module needs to do something in these cases as well, this
        method can be overridden.

        Parameters:
            obj_num:    The object number
            config:     The configuration field for this output object.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress. [default: None]
        """
        pass

    def processImage(self, index, obj_nums, config, base, logger):
        """Perform any necessary processing at the end of each image construction.

        This function will be called after each full image is built.

        Remember, these images may be processed out of order.  But if using the default
        constructor, the data list is already set to be the correct size, so it is safe to
        access self.data[k], where k = base['image_num'] - base['start_image_num'] is the
        appropriate index to use for this image.

        Parameters:
            index:      The index in self.data to use for this image.  This isn't the image_num
                        (which can be accessed at base['image_num'] if needed), but rather
                        an index that starts at 0 for the first image being worked on and
                        goes up to nimages-1.
            obj_nums:   The object numbers that were used for this image.
            config:     The configuration field for this output object.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress. [default: None]
        """
        pass

    def ensureFinalized(self, config, base, main_data, logger):
        """A helper function in the base class to make sure finalize only gets called once by the
        different possible locations that might need it to have been called.

        Parameters:
            config:     The configuration field for this output object.
            base:       The base configuration dict.
            main_data:  The main file data in case it is needed.
            logger:     If given, a logger object to log progress. [default: None]

        Returns:
            the final version of the object.
        """
        if self.final_data is None:
            self.final_data = self.finalize(config, base, main_data, logger)
        return self.final_data

    def finalize(self, config, base, main_data, logger):
        """Perform any final processing at the end of all the image processing.

        This function will be called after all images have been built.

        It returns some sort of final version of the object.  In the base class, it just returns
        self.data, but depending on the meaning of the output object, something else might be
        more appropriate.

        Parameters:
           config:     The configuration field for this output object.
           base:       The base configuration dict.
           main_data:  The main file data in case it is needed.
           logger:     If given, a logger object to log progress. [default: None]

        Returns:
           The final version of the object.
        """
        return self.data

    def writeFile(self, file_name, config, base, logger):
        """Write this output object to a file.

        The base class implementation is appropriate for the cas that the result of finalize
        is a list of images to be written to a FITS file.

        Parameters:
            file_name:  The file to write to.
            config:     The configuration field for this output object.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress. [default: None]
        """
        writeMulti(self.final_data, file_name)

    def writeHdu(self, config, base, logger):
        """Write the data to a FITS HDU with the data for this output object.

        The base class implementation is appropriate for the cas that the result of finalize
        is a list of images of length 1 to be written to a FITS file.

        Parameters:
            config:     The configuration field for this output object.
            base:       The base configuration dict.
            logger:     If given, a logger object to log progress. [default: None]

        Returns:
            an HDU with the output data.
        """
        if len(self.data) != 1:  # pragma: no cover  (Not sure if this is possible.)
            raise GalSimConfigError(
                    "%d %s images were created. Expecting 1."%(n,self._extra_output_key))
        return self.data[0]


def RegisterExtraOutput(key, builder):
    """Register an extra output field for use by the config apparatus.

    The builder parameter should be a subclass of galsim.config.ExtraOutputBuilder.
    See that class for the functions that should be defined and their signatures.
    Not all functions need to be overridden.  If nothing needs to be done at a particular place
    in the processing, you can leave the base class function, which doesn't do anything.

    Parameters:
        key:        The name of the output field in config['output']
        builder:    A builder object to use for building the extra output object.
                    It should be an instance of a subclass of ExtraOutputBuilder.
    """
    builder._extra_output_key = key
    valid_extra_outputs[key] = builder

# Nothing is registered here.  The appropriate items are registered in extra_*.py.
